# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from dataclasses import dataclass
from typing import Any, Generic, List, Optional, TypeVar

from ..cpp import gds_device_buffer
from ..st_types import Device, DType


@dataclass
class TensorBase:
    device: Device
    dtype: DType

    @abstractmethod
    def get_raw(self) -> Any:
        pass

    @abstractmethod
    def contiguous(self) -> "TensorBase":
        pass

    @abstractmethod
    def to(
        self,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> "TensorBase":
        pass

    @abstractmethod
    def clone(self) -> "TensorBase":
        pass

    @abstractmethod
    def detach(self) -> "TensorBase":
        pass

    @abstractmethod
    def view(self, dtype: DType) -> "TensorBase":
        pass

    @abstractmethod
    def __getitem__(self, _val) -> "TensorBase":
        pass


T = TypeVar("T", bound=TensorBase)


class ProcessGroupBase(ABC, Generic[T]):
    @abstractmethod
    def size(self) -> int:
        pass

    @abstractmethod
    def rank(self) -> int:
        pass

    @abstractmethod
    def broadcast(self, dst: T, rank: int) -> None:
        pass

    @abstractmethod
    def scatter(
        self,
        dst: T,
        scatter_list: List[T],
        src: int,
    ) -> None:
        pass

    @abstractmethod
    def send(
        self,
        t: T,
        dst_rank: int,
        tag: int,
    ) -> None:
        pass

    @abstractmethod
    def recv(
        self,
        t: T,
        src_rank: int,
        tag: int,
    ) -> None:
        pass


K = TypeVar("K", bound=ProcessGroupBase)


class FrameworkOpBase(ABC, Generic[T, K]):
    @abstractmethod
    def get_name(self) -> str:
        pass

    @abstractmethod
    def get_device(self, device: str, pg: K) -> Device:
        pass

    @abstractmethod
    def set_device(self, device: Device) -> None:
        pass

    @abstractmethod
    def alloc_tensor_memory(self, length: int, dev: Device) -> gds_device_buffer:
        pass

    @abstractmethod
    def free_tensor_memory(self, gbuf: gds_device_buffer, dev: Device) -> None:
        pass

    @abstractmethod
    def get_empty_tensor(self, shape: List[int], dtype: DType, device: Device) -> T:
        pass

    @abstractmethod
    def concat_tensors(self, tensors: List[T], dim: int) -> T:
        pass

    @abstractmethod
    def copy_tensor(self, dst: T, src: T) -> None:
        pass

    @abstractmethod
    def get_dtype_size(self, dtype: DType) -> int:
        pass

    @abstractmethod
    def from_dlpack(self, dl_tensor: Any, device: Device, dtype: DType) -> T:
        pass

    @abstractmethod
    def get_cuda_ver(self) -> str:
        pass

    @abstractmethod
    def get_device_ptr_align(self) -> int:
        pass

    @abstractmethod
    def as_workaround_dtype(self, dtype: DType) -> DType:
        pass

    @abstractmethod
    def get_process_group(self, pg: Optional[Any]) -> ProcessGroupBase:
        pass

    @abstractmethod
    def is_equal(self, wrapped: T, real: Any) -> bool:
        pass

    @abstractmethod
    def randn(self, s: tuple, device: Device, dtype: DType) -> T:
        pass

    @abstractmethod
    def support_fp8(self) -> bool:
        pass

    @abstractmethod
    def get_mem_used(self) -> int:
        pass


def get_framework_op(name: str) -> FrameworkOpBase:
    if name == "pt" or name == "pytorch" or name == "torch":
        from ._torch import get_framework_op as op
    elif name == "paddle":
        from ._paddle import get_framework_op as op
    else:
        raise Exception(f"Unknown framework name: {name}")
    return op()



# SPDX-License-Identifier: Apache-2.0

import os
import sys
from collections import OrderedDict
from typing import Any, Dict, List, Tuple

import pytest

from fastsafetensors import SafeTensorsFileLoader, SafeTensorsMetadata, SingleGroup
from fastsafetensors import cpp as fstcpp
from fastsafetensors import fastsafe_open
from fastsafetensors.common import get_device_numa_node, is_gpu_found
from fastsafetensors.copier.gds import GdsFileCopier
from fastsafetensors.copier.nogds import NoGdsFileCopier
from fastsafetensors.dlpack import from_cuda_buffer
from fastsafetensors.frameworks import FrameworkOpBase
from fastsafetensors.st_types import Device, DeviceType, DType

# Add tests directory to path to import platform_utils
sys.path.insert(0, os.path.dirname(__file__))
from platform_utils import skip_if_rocm_expected_failure


def load_safetensors_file(
    filename: str,
    device: Device,
    framework: FrameworkOpBase,
    to_dtype: DType = DType.AUTO,
) -> Dict[str, Any]:
    if framework.get_name() == "pytorch":
        from safetensors.torch import load_file
    elif framework.get_name() == "paddle":
        from safetensors.paddle import load_file
    else:
        raise Exception(f"unkown framework: {framework.get_name()}")
    d = load_file(filename, device.as_str())
    if to_dtype != DType.AUTO:
        if framework.get_name() == "pytorch":
            from fastsafetensors.frameworks._torch import dtype_convert
        elif framework.get_name() == "paddle":
            from fastsafetensors.frameworks._paddle import dtype_convert

        for k, t in d.items():
            d[k] = t.to(dtype=dtype_convert[to_dtype])
    return d


def save_safetensors_file(
    tensors: Dict[str, Any],
    filename: str,
    metadata: Dict[str, str],
    framework: FrameworkOpBase,
) -> None:
    if framework.get_name() == "pytorch":
        from safetensors.torch import save_file
    elif framework.get_name() == "paddle":
        from safetensors.paddle import save_file
    else:
        raise Exception(f"unkown framework: {framework.get_name()}")
    save_file(tensors, filename, metadata)


def get_and_check_device(framework: FrameworkOpBase):
    dev_is_gpu = is_gpu_found()
    device = "cpu"
    if dev_is_gpu:
        if framework.get_name() == "pytorch":
            device = "cuda:0"
        elif framework.get_name() == "paddle":
            device = "gpu:0"
    return Device.from_str(device), dev_is_gpu


def run_nogds_file_read(
    input_file: str,
    framework: FrameworkOpBase,
) -> Tuple[SafeTensorsMetadata, fstcpp.gds_device_buffer]:
    fd = os.open(input_file, os.O_RDONLY, 0o644)
    meta = SafeTensorsMetadata.from_file(input_file, framework)
    size = meta.size_bytes - meta.header_length
    device, dev_is_gpu = get_and_check_device(framework)
    gbuf = framework.alloc_tensor_memory(size, device)
    reader = fstcpp.nogds_file_reader(
        False, 20 * 1024, 1, dev_is_gpu, device.index or 0
    )
    req = reader.submit_read(fd, gbuf, meta.header_length, size, 0)
    assert req > 0
    assert reader.wait_read(req) >= 0
    os.close(fd)
    return (meta, gbuf)


def test_device(fstcpp_log) -> None:
    with pytest.raises(ValueError, match="Unknown device type: aaaa"):
        Device.from_str("aaaa:0")
    with pytest.raises(ValueError, match="Invalid index: -xxx"):
        Device.from_str("cpu:-xxx")
    with pytest.raises(ValueError, match="Unknown device type: aaa"):
        Device.from_str("aaa")
    cuda = Device.from_str("cuda:4")
    assert cuda.type == DeviceType.CUDA
    assert cuda.index == 4
    cpu = Device(DeviceType.CPU, None)
    assert cpu.type == DeviceType.CPU and cpu.index == None


def test_framework(fstcpp_log, framework) -> None:
    t = framework.get_empty_tensor([1], DType.F16, Device.from_str("cpu"))
    with pytest.raises(Exception):
        framework.is_equal(t, [float(0.0)])
    with pytest.raises(Exception):
        framework.get_process_group(int(0))
    # Test that get_cuda_ver() returns a string with platform prefix
    cuda_ver = framework.get_cuda_ver()
    assert isinstance(cuda_ver, str)
    # Should be "hip-X.Y.Z", "cuda-X.Y", or "0.0"
    assert (
        cuda_ver.startswith("hip-") or cuda_ver.startswith("cuda-") or cuda_ver == "0.0"
    )

    # Verify it matches what torch reports
    if framework.get_name() == "pytorch":
        import torch

        if torch.cuda.is_available():
            if hasattr(torch.version, "hip") and torch.version.hip:
                assert cuda_ver.startswith("hip-")
                assert str(torch.version.hip) in cuda_ver
            else:
                assert cuda_ver.startswith("cuda-")
                assert str(torch.version.cuda) in cuda_ver
        else:
            assert cuda_ver == "0.0"


def test_get_framework_fail(fstcpp_log) -> None:
    from fastsafetensors.frameworks import get_framework_op

    with pytest.raises(Exception, match="Unknown framework name"):
        get_framework_op("aaaaa")


def make_header_bytes(s: str):
    header = s.encode("utf-8")
    n = len(header)
    return n.to_bytes(8, byteorder="little", signed=False) + header


def test_from_buffer_header_too_small(framework):
    with pytest.raises(Exception, match="HeaderTooSmall"):
        SafeTensorsMetadata.from_buffer(
            buf=0, buffer_len=4, filename="testfile", framework=framework
        )


def test_from_buffer_header_too_large(monkeypatch, framework):
    def fake_read_buffer(buf, size):
        return (100_000_001).to_bytes(8, "little")

    monkeypatch.setattr(fstcpp, "read_buffer", fake_read_buffer)

    with pytest.raises(Exception, match="HeaderTooLarge"):
        SafeTensorsMetadata.from_buffer(
            buf=0, buffer_len=1024, filename="testfile", framework=framework
        )


def test_from_buffer_invalid_header_length(monkeypatch, framework):
    def fake_read_buffer(buf, size):
        return (100).to_bytes(8, "little")

    monkeypatch.setattr(fstcpp, "read_buffer", fake_read_buffer)

    with pytest.raises(Exception, match="InvalidHeaderLength"):
        SafeTensorsMetadata.from_buffer(
            buf=0, buffer_len=50, filename="testfile", framework=framework
        )


def test_from_buffer_success(monkeypatch, framework):
    json_str = '{"__metadata__": {"data_offsets": [0, 123]}}'
    header_bytes = make_header_bytes(json_str)
    buf_data = header_bytes

    def fake_read_buffer(buf, size):
        return buf_data[buf : buf + size]

    monkeypatch.setattr(fstcpp, "read_buffer", fake_read_buffer)

    meta = SafeTensorsMetadata.from_buffer(
        buf=0, buffer_len=len(buf_data), filename="goodfile", framework=framework
    )
    assert isinstance(meta, SafeTensorsMetadata)


def test_load_metadata_and_dlpack(fstcpp_log, input_files, framework) -> None:
    print("test_load_metadata_and_dlpack")
    assert len(input_files) > 0
    device, _ = get_and_check_device(framework)
    for input_file in input_files:
        expected_tensors = load_safetensors_file(input_files[0], device, framework)
        meta, gbuf = run_nogds_file_read(input_file, framework)
        assert meta.header_length > 0
        assert meta.size_bytes > 0
        assert len(meta.tensors) > 0
        printed = False
        for name, actual_meta in sorted(meta.tensors.items(), key=lambda x: x[0]):
            dst_dev_ptr = gbuf.get_base_address() + actual_meta.data_offsets[0]
            wdtype = framework.as_workaround_dtype(actual_meta.dtype)
            cu_buf = from_cuda_buffer(
                dst_dev_ptr, actual_meta.shape, actual_meta.strides, wdtype, device
            )
            actual = framework.from_dlpack(cu_buf, device, wdtype)
            if wdtype != actual_meta.dtype:
                actual = actual.view(actual_meta.dtype)
            exp = expected_tensors[name]
            assert framework.is_equal(actual, exp)
            if not printed:
                print(actual_meta.__repr__())
                printed = True
        framework.free_tensor_memory(gbuf, device)
        assert framework.get_mem_used() == 0
        assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def test_set_debug_log() -> None:
    fstcpp.set_debug_log(False)
    assert True


def test_get_alignment_size() -> None:
    assert fstcpp.get_alignment_size() == 4096


def test_init_gds(fstcpp_log) -> None:
    assert fstcpp.init_gds() == 0


def test_close_gds(fstcpp_log) -> None:
    assert fstcpp.close_gds() == 0


def test_get_device_pci_bus(fstcpp_log) -> None:
    bus = fstcpp.get_device_pci_bus(0)
    if not is_gpu_found():
        assert bus == ""
    else:
        print(f"bus for gpu:0: {bus}")
        assert len(bus) > 0


def test_set_numa_node(fstcpp_log) -> None:
    assert fstcpp.set_numa_node(0) == 0


def test_alloc_gds_buffer(fstcpp_log, framework) -> None:
    print("test_alloc_gds_buffer")
    device, _ = get_and_check_device(framework)
    gbuf = framework.alloc_tensor_memory(1024, device)
    addr = gbuf.get_base_address()
    assert addr != 0
    framework.free_tensor_memory(gbuf, device)


def test_cufile_register_deregister(fstcpp_log, framework) -> None:
    print("test_cufile_register_deregister")
    device, _ = get_and_check_device(framework)
    gbuf = framework.alloc_tensor_memory(1024, device)
    assert gbuf.cufile_register(0, 256) == 0
    assert gbuf.cufile_register(256, 1024 - 256) == 0
    assert gbuf.cufile_deregister(0) == 0
    assert gbuf.cufile_deregister(256) == 0
    framework.free_tensor_memory(gbuf, device)


def test_memmove(fstcpp_log, framework) -> None:
    print("test_memmove")
    device, _ = get_and_check_device(framework)
    gbuf = framework.alloc_tensor_memory(1024, device)
    tmp = framework.alloc_tensor_memory(1024, device)
    assert gbuf.memmove(0, 12, tmp, 256 * 3) == 0
    # Confuse about this test : gbuf.memmove(0, 12, tmp, 1024)
    # I think this test should start copying a section of 1024 memory
    # from the position of gbuf+12 to the position of gbuf+0.
    # However, this piece of memory itself is only 1024.
    # After offsetting by 12, there is no 1024 left in the remaining memory.
    # This part really puzzles me. So I change the moving size to 256*3 (<1024)
    framework.free_tensor_memory(gbuf, device)
    framework.free_tensor_memory(tmp, device)
    assert framework.get_mem_used() == 0


def test_nogds_file_reader(fstcpp_log, input_files, framework) -> None:
    print("test_nogds_file_reader")
    fd = os.open(input_files[0], os.O_RDONLY, 0o644)
    s = os.fstat(fd)
    assert fd > 0
    device, dev_is_gpu = get_and_check_device(framework)
    gbuf = framework.alloc_tensor_memory(s.st_size, device)
    reader = fstcpp.nogds_file_reader(
        False, 256 * 1024, 4, dev_is_gpu, device.index or 0
    )
    step = s.st_size // 4
    reqs = []
    off = 0
    while off < s.st_size:
        l = step
        if off + l > s.st_size:
            l = s.st_size - off
        if 256 * 1024 * 1024 < l:
            l = 256 * 1024 * 1024
        req = reader.submit_read(fd, gbuf, off, l, off)
        assert req > 0
        reqs.append(req)
        off += l
    for req in reqs:
        assert reader.wait_read(req) > 0
    os.close(fd)
    framework.free_tensor_memory(gbuf, device)
    del reader
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def test_NoGdsFileCopier(fstcpp_log, input_files, framework) -> None:
    print("test_NoGdsFileCopier")
    meta = SafeTensorsMetadata.from_file(input_files[0], framework)
    device, dev_is_gpu = get_and_check_device(framework)
    reader = fstcpp.nogds_file_reader(
        False, 256 * 1024, 4, dev_is_gpu, device.index or 0
    )
    copier = NoGdsFileCopier(meta, device, reader, framework)
    gbuf = copier.submit_io(False, 10 * 1024 * 1024 * 1024)
    tensors = copier.wait_io(gbuf)
    for key, exp in load_safetensors_file(input_files[0], device, framework).items():
        actual = tensors[key]
        assert framework.is_equal(actual, exp)
    framework.free_tensor_memory(gbuf, device)
    del copier
    del reader
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def test_GdsFileCopier(fstcpp_log, input_files, framework) -> None:
    print("test_GdsFileCopier")
    skip_if_rocm_expected_failure("test_GdsFileCopier")
    meta = SafeTensorsMetadata.from_file(input_files[0], framework)
    device, dev_is_gpu = get_and_check_device(framework)
    reader = fstcpp.gds_file_reader(4, dev_is_gpu, device.index or 0)
    copier = GdsFileCopier(meta, device, reader, framework)
    gbuf = copier.submit_io(False, 10 * 1024 * 1024 * 1024)
    tensors = copier.wait_io(gbuf)
    for key, exp in load_safetensors_file(input_files[0], device, framework).items():
        actual = tensors[key]
        assert framework.is_equal(actual, exp)
    framework.free_tensor_memory(gbuf, device)
    del reader
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def test_SafeTensorsFileLoader(fstcpp_log, input_files, framework) -> None:
    device, _ = get_and_check_device(framework)
    if framework.get_name() == "pytorch":
        import torch

        data_type = DType.F16
        data_type_real = torch.float16
    elif framework.get_name() == "paddle":
        # There are some lack of accuracy in paddle.float16 (about 1e-4) in cpu.
        import paddle

        data_type = DType.F32
        data_type_real = paddle.float32
    else:
        raise NotImplementedError(
            f"Do not support the framework: {framework.get_name()}"
        )
    loader = SafeTensorsFileLoader(
        pg=SingleGroup(),
        device=device.as_str(),
        framework=framework.get_name(),
        nogds=False,
        debug_log=True,
    )
    loader.add_filenames({0: input_files})
    bufs = loader.copy_files_to_device(
        dtype=data_type, use_buf_register=True, max_copy_block_size=256 * 1024 * 1024
    )
    last_key = ""
    last_shape: List[int] = []
    for key, exp in load_safetensors_file(input_files[0], device, framework).items():
        exp = exp.to(dtype=data_type_real)
        actual = bufs.get_tensor_wrapped(key)
        assert framework.is_equal(actual, exp)
        last_key = key
        last_shape = list(exp.shape)
    if last_key != "":
        assert bufs.get_filename(last_key) == input_files[0]
        assert bufs.get_shape(last_key) == last_shape
        assert loader.get_shape(last_key) == last_shape
    assert bufs.get_filename("aaaaaaaaaaaaa") == ""
    bufs.close()
    loader.close()
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def test_SafeTensorsFileLoaderNoGds(fstcpp_log, input_files, framework) -> None:
    device, _ = get_and_check_device(framework)
    loader = SafeTensorsFileLoader(
        pg=SingleGroup(),
        device=device.as_str(),
        framework=framework.get_name(),
        nogds=True,
        debug_log=True,
    )
    loader.add_filenames({0: input_files})
    bufs = loader.copy_files_to_device()
    key_dims = OrderedDict({key: -1 for key in loader.get_keys()})
    tensors = bufs.as_dict(key_dims)
    for key, exp in load_safetensors_file(input_files[0], device, framework).items():
        actual = tensors[key]
        assert framework.is_equal(actual, exp)
    bufs.close()
    loader.close()
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def test_fastsafe_open(fstcpp_log, input_files, framework) -> None:
    device, _ = get_and_check_device(framework)

    def weight_iterator():
        with fastsafe_open(
            input_files,
            device=device.as_str(),
            nogds=True,
            debug_log=True,
            framework=framework.get_name(),
        ) as f:
            for k in f.keys():
                t = f.get_tensor_wrapped(k)
                yield k, t

    tensors = load_safetensors_file(input_files[0], device, framework)
    for k, t in weight_iterator():
        assert framework.is_equal(t, tensors[k])

    with fastsafe_open(
        input_files[0],
        device=device.as_str(),
        nogds=True,
        framework=framework.get_name(),
    ) as f:
        for filename in f.metadata().keys():
            assert filename in input_files

    with fastsafe_open(
        {0: input_files},
        device=device.as_str(),
        nogds=True,
        framework=framework.get_name(),
    ) as f:
        for k in f.keys():
            t = f.get_tensor(k)
            if framework.get_name() == "pytorch":
                import torch

                assert isinstance(t, torch.Tensor)
            elif framework.get_name() == "paddle":
                import paddle

                assert isinstance(t, paddle.Tensor)
            break
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def _test_type(
    tmp_dir,
    dtype: DType,
    device: Device,
    framework: FrameworkOpBase,
    to_dtype: DType = DType.AUTO,
) -> None:
    filename = os.path.join(tmp_dir, f"a.safetensors")
    t0 = framework.randn((8, 16), device=device, dtype=DType.F32).to(dtype=dtype)
    if to_dtype is not DType.AUTO:
        t0 = t0.to(dtype=to_dtype)
    save_safetensors_file({f"a": t0.get_raw()}, filename, {"fst": "sample"}, framework)
    t2 = load_safetensors_file(filename, device, framework, to_dtype=to_dtype)
    with fastsafe_open(
        filenames=[filename],
        nogds=True,
        device=device.as_str(),
        framework=framework.get_name(),
        debug_log=True,
    ) as f:
        for key in f.keys():
            t1 = f.get_tensor_wrapped(key).clone().detach()
            assert framework.is_equal(t1, t2[key])
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


def test_int8(fstcpp_log, tmp_dir, framework) -> None:
    if not framework.support_fp8():
        pytest.skip("FP8 is not supported")
        return
    device, _ = get_and_check_device(framework)
    _test_type(tmp_dir, DType.I8, device, framework)


def test_float8_e5m2(fstcpp_log, tmp_dir, framework) -> None:
    if not framework.support_fp8():
        pytest.skip("FP8 is not supported")
        return
    device, _ = get_and_check_device(framework)
    _test_type(tmp_dir, DType.F8_E5M2, device, framework)


def test_float8_e4m3fn(fstcpp_log, tmp_dir, framework) -> None:
    if not framework.support_fp8():
        pytest.skip("FP8 is not supported")
        return
    device, _ = get_and_check_device(framework)
    _test_type(tmp_dir, DType.F8_E4M3, device, framework)


def test_float8_e4m3fn_to_int8(fstcpp_log, tmp_dir, framework) -> None:
    if not framework.support_fp8():
        pytest.skip("FP8 is not supported")
        return
    device, _ = get_and_check_device(framework)
    _test_type(tmp_dir, DType.F8_E4M3, device, framework, DType.I8)


def test_cpp_metrics(fstcpp_log, framework) -> None:
    device, _ = get_and_check_device(framework)
    exp_length = 0
    assert framework.get_mem_used() == exp_length

    gbuf = framework.alloc_tensor_memory(128, device)
    exp_length += 128
    assert framework.get_mem_used() == exp_length

    framework.free_tensor_memory(gbuf, device)
    exp_length -= 128
    assert framework.get_mem_used() == exp_length

    gbuf2 = framework.alloc_tensor_memory(1024, device)
    exp_length += 1024
    assert framework.get_mem_used() == exp_length

    gbuf3 = framework.alloc_tensor_memory(128, device)
    exp_length += 128
    assert framework.get_mem_used() == exp_length

    framework.free_tensor_memory(gbuf2, device)
    exp_length -= 1024
    assert framework.get_mem_used() == exp_length

    framework.free_tensor_memory(gbuf3, device)
    exp_length -= 128
    assert framework.get_mem_used() == exp_length

    assert exp_length == 0
    assert framework.get_mem_used() == 0



# SPDX-License-Identifier: Apache-2.0

from dataclasses import dataclass
from enum import Enum
from typing import Optional


class DeviceType(Enum):
    CPU = "cpu"
    CUDA = "cuda"
    GPU = "gpu"


@dataclass(frozen=True)
class Device:
    type: DeviceType = DeviceType.CPU
    index: Optional[int] = None

    def as_str(self) -> str:
        if self.index is None:
            return self.type.value
        return f"{self.type.value}:{self.index}"

    @classmethod
    def from_str(cls, s: str) -> "Device":
        if ":" in s:
            type_str, index_str = s.split(":", 1)
            try:
                dev_type = DeviceType(type_str.lower())
            except ValueError:
                raise ValueError(f"Unknown device type: {type_str}")
            try:
                index = int(index_str)
            except ValueError:
                raise ValueError(f"Invalid index: {index_str}")
            return cls(type=dev_type, index=index)
        else:
            try:
                dev_type = DeviceType(s.lower())
            except ValueError:
                raise ValueError(f"Unknown device type: {s}")
            return cls(type=dev_type, index=None)


class DType(Enum):
    BOOL = "BOOL"
    I8 = "I8"
    I16 = "I16"
    I32 = "I32"
    I64 = "I64"
    U8 = "U8"
    U16 = "U16"
    U32 = "U32"
    U64 = "U64"
    F16 = "F16"
    F32 = "F32"
    F64 = "F64"
    BF16 = "BF16"
    F8_E5M2 = "F8_E5M2"
    F8_E4M3 = "F8_E4M3"
    AUTO = "AUTO"



# SPDX-License-Identifier: Apache-2.0

import os
from typing import Dict, List

from .. import cpp as fstcpp
from ..common import SafeTensorsMetadata
from ..frameworks import FrameworkOpBase, TensorBase
from ..st_types import Device, DType
from .base import CopierInterface


class NoGdsFileCopier(CopierInterface):
    def __init__(
        self,
        metadata: SafeTensorsMetadata,
        device: Device,
        reader: fstcpp.nogds_file_reader,
        framework: FrameworkOpBase,
    ):
        self.framework = framework
        self.metadata = metadata
        self.reader = reader
        self.fd = os.open(metadata.src, os.O_RDONLY, 0o644)
        if self.fd < 0:
            raise Exception(
                f"NoGdsFileCopier.__init__: failed to open, file={metadata.src}"
            )
        self.device = device
        self.reqs: List[int] = []

    def submit_io(
        self, use_buf_register: bool, max_copy_block_size: int
    ) -> fstcpp.gds_device_buffer:
        total_length = self.metadata.size_bytes - self.metadata.header_length
        gbuf = self.framework.alloc_tensor_memory(total_length, self.device)
        count = 0
        while count < total_length:
            l = total_length - count
            if max_copy_block_size < l:
                l = max_copy_block_size
            req = self.reader.submit_read(
                self.fd, gbuf, self.metadata.header_length + count, l, count
            )
            if req < 0:
                raise Exception(f"submit_io: submit_nogds_read failed, err={req}")
            self.reqs.append(req)
            count += l
        return gbuf

    def wait_io(
        self,
        gbuf: fstcpp.gds_device_buffer,
        dtype: DType = DType.AUTO,
        noalign: bool = False,
    ) -> Dict[str, TensorBase]:
        for req in self.reqs:
            count = self.reader.wait_read(req)
            if count < 0:
                raise Exception(f"wait_io: wait_nogds_read failed, req={req}")
        if self.fd > 0:
            os.close(self.fd)
            self.fd = 0
        return self.metadata.get_tensors(
            gbuf, self.device, self.metadata.header_length, dtype=dtype
        )



# SPDX-License-Identifier: Apache-2.0

from typing import Any, Dict

from .. import cpp as fstcpp
from ..common import SafeTensorsMetadata
from ..frameworks import FrameworkOpBase, TensorBase
from ..st_types import Device, DeviceType, DType
from .base import CopierInterface, DummyDeviceBuffer


class ExampleCopier(CopierInterface):
    def __init__(
        self,
        metadata: SafeTensorsMetadata,
        device: Device,
        reader,
        framework: FrameworkOpBase,
    ):
        pass

    def submit_io(
        self, use_buf_register: bool, max_copy_block_size: int
    ) -> fstcpp.gds_device_buffer:
        return DummyDeviceBuffer()

    def wait_io(
        self,
        gbuf: fstcpp.gds_device_buffer,
        dtype: DType = DType.AUTO,
        noalign: bool = False,
    ) -> Dict[str, TensorBase]:
        # get tensor
        res: Dict[str, TensorBase] = {}
        return res


def new_gds_file_copier(
    device: Device,
    bbuf_size_kb: int = 16 * 1024,
    max_threads: int = 16,
    nogds: bool = False,
):
    # reader = example_reader()
    reader: Any = {}

    def construct_copier(
        metadata: SafeTensorsMetadata,
        device: Device,
        framework: FrameworkOpBase,
    ) -> CopierInterface:
        return ExampleCopier(metadata, device, reader, framework)

    return construct_copier



# SPDX-License-Identifier: Apache-2.0

import json
import logging
import os
import sys
from collections import OrderedDict
from dataclasses import dataclass
from typing import Dict, List, Optional, Tuple

from . import cpp as fstcpp
from .dlpack import from_cuda_buffer
from .frameworks import FrameworkOpBase, TensorBase
from .st_types import Device, DType


def init_logger(name: str):
    return logging.getLogger(name)


def set_debug():
    logging.basicConfig(
        format="[%(levelname)s] %(message)s", level=logging.DEBUG, force=True
    )


def is_debug(logger: logging.Logger) -> bool:
    return logger.isEnabledFor(logging.DEBUG)


def is_gpu_found():
    """Check if any GPU (CUDA or HIP) is available.

    Returns True if either CUDA or ROCm/HIP GPUs are detected.
    This allows code to work transparently across both platforms.
    """
    return fstcpp.is_cuda_found() or fstcpp.is_hip_found()


def get_device_numa_node(device: Optional[int]) -> Optional[int]:
    if device is None or not sys.platform.startswith("linux"):
        return None
    pci_addr = fstcpp.get_device_pci_bus(device)
    if pci_addr == "":
        return None
    bus_addr = ":".join(pci_addr.split(":")[:2]).lower()
    syspath = f"/sys/class/pci_bus/{bus_addr}/device/numa_node"
    if not os.path.exists(syspath):
        return None
    with open(syspath) as f:
        return int(f.read().strip())


# keep this for compatibility
class SingleGroup:
    def size(self):
        return 1

    def rank(self):
        return 0


class SafeTensorsMetadata:
    def __init__(
        self,
        string: str,
        header_length: int,
        size_bytes: int,
        framework: FrameworkOpBase,
        src: str = "",
        keep_orig_dict: bool = False,
    ):
        self.src = src
        self.framework = framework
        ser = json.loads(string, object_pairs_hook=OrderedDict)
        self.metadata = ser.get("__metadata__", "")
        if self.metadata:
            del ser["__metadata__"]
        self.tensors: Dict[str, TensorFrame] = {}
        self.header_length = header_length
        self.aligned = header_length % framework.get_device_ptr_align() == 0
        if keep_orig_dict:
            self.ser = ser

        start = 0
        for _, (k, buffer) in enumerate(
            sorted(ser.items(), key=lambda x: x[1]["data_offsets"][0])
        ):
            t: TensorFrame = TensorFrame.from_buffer(buffer)
            self.tensors[k] = t
            # validation
            s, e = t.data_offsets
            if s != start or e < s:
                raise Exception(
                    f"validate(tensor {k}): InvalidOffset s={s}, start={start}, e={e}, src={src}"
                )
            # if (header_length + s) % CUDA_PTR_ALIGN > 0:
            #    logger.warning(f"misaligned tensor is detected at {header_length + s}. this will cause cuda pointer alignment errors later.")
            start = e
            nelements = 1
            for sh in t.shape:
                nelements *= sh
            nbytes = nelements * framework.get_dtype_size(t.dtype)
            if (e - s) != nbytes:
                raise Exception(
                    f"validate(tensor {k}): TensorInvalidInfo, e-s={e-s}, nbytes={nbytes}, src={src}"
                )
        self.size_bytes = size_bytes
        if start + header_length != size_bytes:
            raise Exception(
                f"MetadataIncompleteBuffer, src={src}, start={start}, header_length={header_length}, size_bytes={size_bytes}"
            )

    @classmethod
    def from_buffer(
        self, buf: int, buffer_len: int, filename: str, framework: FrameworkOpBase
    ):
        if buffer_len < 8:
            raise Exception(
                f"from_buffer: HeaderTooSmall, filename={filename}, buffer_len={buffer_len}"
            )
        arr = fstcpp.read_buffer(buf, 8)
        n = int.from_bytes(arr, byteorder="little", signed=False)
        if n > 100000000:
            raise Exception(
                f"from_buffer: HeaderTooLarge, n={n}, filename={filename}, buffer_len={buffer_len}"
            )
        if n > buffer_len - 8:
            raise Exception(
                f"from_buffer: InvalidHeaderLength, n={n}, filename={filename}, buffer_len={buffer_len}"
            )
        string = fstcpp.read_buffer(buf + 8, n).decode("utf-8")
        # Assert the string starts with {
        # NOTE: Add when we move to 0.4.0
        # if string.startswith('{'):
        #    raise Exception(f"{filename}: InvalidHeaderStart")
        return SafeTensorsMetadata(string, n + 8, buffer_len, framework)

    @classmethod
    def from_fd(
        self,
        fd: int,
        filename: str,
        framework: FrameworkOpBase,
        keep_orig_dict: bool = False,
    ):
        status = os.fstat(fd)
        buffer_len = status.st_size
        if buffer_len < 8:
            raise Exception(f"{filename}: HeaderTooSmall, buffer_len={buffer_len}")
        arr = os.read(fd, 8)
        n = int.from_bytes(arr, byteorder="little", signed=False)
        if n > 100000000:
            raise Exception(
                f"{filename}: HeaderTooLarge, n={n}, buffer_len={buffer_len}"
            )
        if n > buffer_len - 8:
            raise Exception(
                f"{filename}: InvalidHeaderLength, n={n}, buffer_len={buffer_len}"
            )
        string = os.read(fd, n).decode("utf-8")
        # Assert the string starts with {
        # NOTE: Add when we move to 0.4.0
        # if string.startswith('{'):
        #    raise Exception(f"{filename}: InvalidHeaderStart")
        return SafeTensorsMetadata(
            string,
            n + 8,
            buffer_len,
            framework,
            filename,
            keep_orig_dict=keep_orig_dict,
        )

    @classmethod
    def from_file(self, filename: str, framework: FrameworkOpBase):
        fd = os.open(filename, os.O_RDONLY, 0o644)
        ret = self.from_fd(fd, filename, framework=framework, keep_orig_dict=False)
        os.close(fd)
        return ret

    def get_tensors(
        self,
        gbuf: fstcpp.gds_device_buffer,
        device: Device,
        copy_start_offset: int,
        dtype: DType = DType.AUTO,
    ) -> Dict[str, TensorBase]:
        ret = {}
        for tensor_name, t in self.tensors.items():
            dst_dev_ptr = (
                gbuf.get_base_address()
                + self.header_length
                + t.data_offsets[0]
                - copy_start_offset
            )
            disk_dtype = self.framework.as_workaround_dtype(t.dtype)
            dl_tensor = from_cuda_buffer(
                dst_dev_ptr,
                t.shape,
                t.strides,
                disk_dtype,
                device,
            )
            t2 = self.framework.from_dlpack(dl_tensor, device, disk_dtype)
            if disk_dtype != t.dtype:
                t2 = t2.view(t.dtype)

            if dtype != DType.AUTO and dtype != t.dtype:
                if self.framework.get_dtype_size(dtype) > self.framework.get_dtype_size(
                    t.dtype
                ):
                    raise Exception(
                        f"Online type conversion to larger sizes is not supported ({t.dtype} -> {dtype})"
                    )
                t3 = t2.to(dtype=dtype)
                conv_dtype: DType = self.framework.as_workaround_dtype(dtype)
                dl_tensor = from_cuda_buffer(
                    dst_dev_ptr,
                    t.shape,
                    t.strides,
                    conv_dtype,
                    device,
                )
                t2 = self.framework.from_dlpack(dl_tensor, device, conv_dtype)
                if dtype != conv_dtype:
                    t2 = t2.view(dtype)
                self.framework.copy_tensor(t2, t3)
                self.tensors[tensor_name].dtype = dtype
            ret[tensor_name] = t2
        return ret

    def __repr__(self) -> str:
        return str({"__metadata__": self.metadata, "tensors": self.tensors})


@dataclass
class TensorFrame:
    dtype: DType
    shape: List[int]
    data_offsets: List[int]
    strides: List[int]
    offsets: List[int]
    sliced: bool

    @classmethod
    def from_buffer(self, entry: OrderedDict[str, List[int]]):
        shape = entry["shape"]
        data_offsets = list(entry["data_offsets"])
        strides = []
        offsets = []
        for i in range(0, len(shape)):
            s = 1
            for j in range(i + 1, len(shape)):
                s *= shape[j]
            strides.append(s)
            offsets.append(0)
        return TensorFrame(
            DType(entry["dtype"]), shape, data_offsets, strides, offsets, False
        )

    def __repr__(self) -> str:
        return str(
            {
                "dtype": self.dtype,
                "shape": self.shape,
                "data_offsets": self.data_offsets,
            }
        )

    # TODO: reduce dim if isinstance(_val, int) == True
    def __getitem__(self, _val) -> "TensorFrame":
        val: Tuple = ()
        if isinstance(_val, slice) or isinstance(_val, int):
            val = (_val,)
        elif isinstance(_val, tuple):
            val = _val
        else:
            raise Exception(f"[BUG] Unsupported index type for DiskTensor: {_val}")
        if len(val) > len(self.shape):
            raise Exception(
                f"[BUG] tried to get too large slice {_val} from {self.shape}"
            )
        shape: List[int] = []
        strides: List[int] = []
        offsets: List[int] = []
        for dim in range(0, len(val)):
            if isinstance(val[dim], int):
                start = val[dim]
                if start >= self.shape[dim] or start < -self.shape[dim]:
                    raise IndexError(
                        f"[BUG] tried to access index {start} at dim={dim} for shape={self.shape}"
                    )
                if start < 0:
                    start = self.shape[dim] + start + 1
                stop = start + 1
                step = 1
                length = 1
            elif isinstance(val[dim], slice):
                start = val[dim].start
                if start is None:
                    start = 0
                if start >= self.shape[dim] or start < -self.shape[dim]:
                    start = self.shape[dim]
                if start < 0:
                    start = self.shape[dim] + start + 1
                stop = val[dim].stop
                if stop is None or stop >= self.shape[dim] or stop < -self.shape[dim]:
                    stop = self.shape[dim]
                if stop < 0:
                    stop = self.shape[dim] + stop + 1
                step = val[dim].step
                if step is None:
                    step = 1
                if step == 0:
                    raise ValueError(f"[BUG] slice step cannot be zero")
                length = stop - start
                if (
                    length == 0
                    or (length < 0 and step > 0)
                    or (length > 0 and step < 0)
                ):
                    return TensorFrame(self.dtype, [], self.data_offsets, [], [], False)
                if length < 0 and step < 0:
                    length *= -1
            else:
                raise Exception(
                    f"[BUG] Unsupported index type for DiskTensor: {_val} at dim={dim}"
                )
            offsets.append(self.offsets[dim] + start)
            strides.append(self.strides[dim] * step)
            shape.append(length // (step if step > 0 else -step))
        for rdim in range(dim + 1, len(self.shape)):
            offsets.append(self.offsets[rdim])
            strides.append(self.strides[rdim])
            shape.append(self.shape[rdim])
        return TensorFrame(self.dtype, shape, self.data_offsets, strides, offsets, True)



# SPDX-License-Identifier: Apache-2.0

try:
    import torch
    import torch.distributed as dist
except ImportError as e:
    raise ImportError("could not import torch. Please install it.") from e

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from ..common import SingleGroup
from ..cpp import cpu_free, cpu_malloc, gds_device_buffer
from ..st_types import Device, DeviceType, DType
from . import FrameworkOpBase, ProcessGroupBase, TensorBase

dtype_convert: Dict[DType, Any] = {
    DType.BOOL: torch.bool,
    DType.U8: torch.uint8,
    DType.I8: torch.int8,
    DType.I16: torch.int16,
    DType.I32: torch.int32,
    DType.I64: torch.int64,
    DType.F16: torch.float16,
    DType.BF16: torch.bfloat16,
    DType.F32: torch.float32,
    DType.F64: torch.float64,
}
need_workaround_dtypes: Dict[DType, DType] = {
    DType.F8_E5M2: DType.I8,
    DType.F8_E4M3: DType.I8,
}

if hasattr(torch, "float8_e5m2"):
    dtype_convert[DType.F8_E5M2] = torch.float8_e5m2
if hasattr(torch, "float8_e4m3fn"):
    dtype_convert[DType.F8_E4M3] = torch.float8_e4m3fn
if hasattr(torch, "uint16"):
    dtype_convert[DType.U16] = torch.uint16
if hasattr(torch, "uint32"):
    dtype_convert[DType.U32] = torch.uint32
if hasattr(torch, "uint64"):
    dtype_convert[DType.U64] = torch.uint64


@dataclass
class TorchTensor(TensorBase):
    real_tensor: torch.Tensor

    def get_raw(self) -> torch.Tensor:
        return self.real_tensor

    def contiguous(self) -> "TorchTensor":
        return TorchTensor(self.device, self.dtype, self.real_tensor.contiguous())

    def to(
        self,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> "TorchTensor":
        to_dev: Optional[str] = None
        if device is not None and self.device != device:
            to_dev = device.as_str()
        else:
            device = self.device
        to_dtype: Optional[torch.dtype] = None
        if dtype != DType.AUTO and (dtype != self.dtype):
            to_dtype = dtype_convert[dtype]
        else:
            dtype = self.dtype
        if to_dev is not None or to_dtype is not None:
            return TorchTensor(
                device, dtype, self.real_tensor.to(device=to_dev, dtype=to_dtype)
            )
        return self

    def clone(self) -> "TorchTensor":
        return TorchTensor(self.device, self.dtype, self.real_tensor.clone())

    def detach(self) -> "TorchTensor":
        return TorchTensor(self.device, self.dtype, self.real_tensor.detach())

    def view(self, dtype: DType) -> "TorchTensor":
        t2 = self.real_tensor.view(dtype_convert[dtype])
        return TorchTensor(self.device, dtype, t2)

    def __getitem__(self, _val) -> "TorchTensor":
        return TorchTensor(self.device, self.dtype, self.real_tensor[_val])


@dataclass
class TorchProcessGroup(ProcessGroupBase[TorchTensor]):
    real_pg: Optional[dist.ProcessGroup]

    def size(self) -> int:
        return self.real_pg.size() if self.real_pg else 1

    def rank(self) -> int:
        return self.real_pg.rank() if self.real_pg else 0

    def broadcast(self, dst: TorchTensor, rank: int) -> None:
        if self.real_pg:
            dist.broadcast(dst.real_tensor, rank, group=self.real_pg)
            # Synchronize to ensure the NCCL broadcast (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if torch.cuda.is_available():
                torch.cuda.synchronize()

    def scatter(
        self,
        dst: TorchTensor,
        scatter_list: List[TorchTensor],
        src: int,
    ) -> None:
        if self.real_pg:
            sl = [t.real_tensor for t in scatter_list]
            dist.scatter(dst.real_tensor, scatter_list=sl, src=src, group=self.real_pg)
            # Synchronize to ensure the NCCL scatter (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if torch.cuda.is_available():
                torch.cuda.synchronize()

    def send(
        self,
        t: TorchTensor,
        dst_rank: int,
        tag: int,
    ):
        if self.real_pg:
            dist.send(t.real_tensor, dst_rank, group=self.real_pg, tag=tag)
            # Synchronize to ensure the NCCL scatter (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if torch.cuda.is_available():
                torch.cuda.synchronize()

    def recv(
        self,
        t: TorchTensor,
        src_rank: int,
        tag: int,
    ):
        if self.real_pg:
            dist.recv(t.real_tensor, src_rank, group=self.real_pg, tag=tag)
            # Synchronize to ensure the NCCL recv (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if torch.cuda.is_available():
                torch.cuda.synchronize()


class TorchOp(FrameworkOpBase[TorchTensor, TorchProcessGroup]):
    def __init__(self) -> None:
        self.mem_used = 0

    def get_name(self) -> str:
        return "pytorch"

    def get_device(self, device: str, pg: TorchProcessGroup) -> Device:
        dev = torch.device(device)
        return Device(DeviceType(dev.type), dev.index)

    def set_device(self, device: Device) -> None:
        if device.type != DeviceType.CPU:
            torch.cuda.set_device(device.as_str())

    def alloc_tensor_memory(self, length: int, dev: Device) -> gds_device_buffer:
        if dev.type == DeviceType.CUDA:
            rbuf = torch.cuda.caching_allocator_alloc(length)
        else:
            rbuf = cpu_malloc(length)
        self.mem_used += length
        return gds_device_buffer(rbuf, length, dev.type == DeviceType.CUDA)

    def free_tensor_memory(self, gbuf: gds_device_buffer, dev: Device):
        length = gbuf.get_length()
        if dev.type == DeviceType.CUDA:
            torch.cuda.caching_allocator_delete(gbuf.get_base_address())
        else:
            cpu_free(gbuf.get_base_address())
        self.mem_used -= length

    def get_empty_tensor(
        self, shape: List[int], dtype: DType, device: Device
    ) -> TorchTensor:
        dst = torch.empty(
            size=shape, dtype=dtype_convert[dtype], device=device.as_str()
        )
        return TorchTensor(device, dtype, dst)

    def concat_tensors(self, tensors: List[TorchTensor], dim: int) -> TorchTensor:
        ts = [tensor.real_tensor for tensor in tensors]
        return TorchTensor(tensors[0].device, tensors[0].dtype, torch.cat(ts, dim=dim))

    def get_dtype_size(self, dtype: DType) -> int:
        return dtype_convert[dtype].itemsize

    def from_dlpack(self, dl_tensor: Any, device: Device, dtype: DType) -> TorchTensor:
        t = torch.from_dlpack(dl_tensor)
        return TorchTensor(device, dtype, t)

    def copy_tensor(self, dst: TorchTensor, src: TorchTensor):
        dst.real_tensor.copy_(src.real_tensor)

    def get_cuda_ver(self) -> str:
        """Get GPU runtime version with platform indicator.

        Returns a string like 'hip-5.7.0' for ROCm or 'cuda-12.1' for CUDA,
        or '0.0' if no GPU is available. This allows code to distinguish
        between different GPU platforms without using torch directly.
        """
        if torch.cuda.is_available():
            # Check if this is ROCm/HIP build
            if hasattr(torch.version, "hip") and torch.version.hip is not None:
                return f"hip-{torch.version.hip}"
            return f"cuda-{torch.version.cuda}"
        return "0.0"

    def get_device_ptr_align(self) -> int:
        CUDA_PTR_ALIGN: int = 16
        return CUDA_PTR_ALIGN

    def as_workaround_dtype(self, dtype: DType) -> DType:
        if dtype in need_workaround_dtypes:
            return need_workaround_dtypes[dtype]
        return dtype

    def get_process_group(self, pg: Optional[Any]) -> TorchProcessGroup:
        if pg is not None:
            if isinstance(pg, SingleGroup):
                pg = None
            elif not isinstance(pg, dist.ProcessGroup):
                raise Exception(
                    "pg must be an instance of torch.disributed.ProcessGroup"
                )
        return TorchProcessGroup(pg)

    # for testing
    def is_equal(self, wrapped: TorchTensor, real: Any) -> bool:
        if isinstance(real, torch.Tensor):
            return bool(torch.all(wrapped.real_tensor.eq(real)))
        raise Exception("real is not torch.Tensor")

    def randn(self, s: tuple, device: Device, dtype: DType) -> TorchTensor:
        return TorchTensor(
            device,
            dtype,
            torch.randn(s, device=device.as_str(), dtype=dtype_convert[dtype]),
        )

    def support_fp8(self) -> bool:
        return DType.F8_E5M2 in dtype_convert

    def get_mem_used(self):
        return self.mem_used


_op: Optional[TorchOp] = None


def get_framework_op() -> FrameworkOpBase:
    global _op
    if _op is None:
        _op = TorchOp()
    return _op



# SPDX-License-Identifier: Apache-2.0

import os

import pytest

os.environ["VLLM_WORKER_MULTIPROC_METHOD"] = "spawn"
import vllm
from vllm import LLM, SamplingParams

sampling_params = SamplingParams(
    temperature=0.0,
    top_p=0.95,
    max_tokens=16,
)

prompts = [
    "Explain the difference between CPU and GPU.",
    "Write a haiku about programming.",
]


def run_llm(llm):
    outputs = llm.generate(prompts, sampling_params)
    for i, output in enumerate(outputs):
        print("=========")
        print(f"Prompt: {prompts[i]}")
        print(f"Output: {output.outputs[0].text}")


def test_vllm_no_fastsafetensors(fstcpp_log):
    run_llm(LLM(model="ibm-granite/granite-3.3-2b-instruct", max_model_len=16))


def test_vllm_fastsafetensors(fstcpp_log):
    run_llm(
        LLM(
            model="ibm-granite/granite-3.3-2b-instruct",
            load_format="fastsafetensors",
            max_model_len=16,
        )
    )


def test_deepseek_r1(fstcpp_log):
    run_llm(
        LLM(
            model="silence09/DeepSeek-R1-Small-2layers",
            load_format="fastsafetensors",
            max_model_len=16,
        )
    )



# SPDX-License-Identifier: Apache-2.0

import pytest

from fastsafetensors import SafeTensorsFileLoader
from fastsafetensors import cpp as fstcpp
from fastsafetensors.common import is_gpu_found


def test_shuffle(fstcpp_log, input_files, pg, framework):
    print("test_shuffle")
    if framework.get_name() == "pytorch":
        from safetensors.torch import load_file

        rank = pg.rank()
        world_size = pg.size()
        device = "cuda:0" if is_gpu_found() else "cpu"
    elif framework.get_name() == "paddle":
        from safetensors.paddle import load_file

        rank = pg.process_group.rank()
        world_size = pg.process_group.size()
        device = "gpu:0" if is_gpu_found() else "cpu"
    else:
        raise Exception(f"Unknown framework: {framework.get_name()}")
    loader = SafeTensorsFileLoader(
        pg=pg, device=device, nogds=True, framework=framework.get_name(), debug_log=True
    )
    loader.add_filenames({0: input_files})
    bufs = loader.copy_files_to_device()
    key_dims = {key: -1 for key in loader.get_keys()}
    for i in range(0, 12):
        key_dims[f"h.{i}.mlp.c_proj.weight"] = 1
        key_dims[f"h.{i}.mlp.c_fc.weight"] = 0
        key_dims[f"h.{i}.attn.c_proj.weight"] = 0
    tensors = bufs.as_dict(key_dims)
    f = load_file(input_files[0])
    origs = {}
    for key in tensors.keys():
        dim = key_dims[key]
        if dim == 0 or dim == 1:
            t = f[key]
            rank_slices = ()
            shape = t.shape
            size = shape[dim]
            block_size = (size + world_size - 1) // world_size
            for i in range(0, len(shape)):
                if i < dim:
                    rank_slices += (slice(None, None, None),)
                elif i == dim:
                    rank_slices += (
                        slice(rank * block_size, (rank + 1) * block_size, 1),
                    )
                    break
            t = t[rank_slices]
            t = t.clone().detach()
        else:
            t = f[key]
        t = t.to(device=device)
        origs[key] = t
        assert framework.is_equal(tensors[key], t)

    bufs.close()
    loader.reset()

    loader.add_filenames({0: input_files})
    bufs = loader.copy_files_to_device()
    if world_size > 1:
        pushed = bufs.push_tensor("h.0.attn.c_proj.bias", 1)
        assert pushed is not None if rank == 1 else pushed is None
        pushed1 = bufs.push_tensor("h.1.attn.c_proj.bias", 0)
        assert pushed1 is not None if rank == 0 else pushed1 is None

        tensors2 = bufs.get_tensor_wrapped("h.0.attn.c_proj.bias")  # cached load
        assert (
            framework.is_equal(tensors2, origs["h.0.attn.c_proj.bias"])
            if rank == 1
            else pushed is None
        )

    actual = bufs.get_multi_cols(
        ["h.0.attn.c_proj.weight", "h.1.attn.c_proj.weight"], dim=0
    )
    if framework.get_name() == "pytorch":
        import torch

        exp = torch.concat(
            [origs["h.0.attn.c_proj.weight"], origs["h.1.attn.c_proj.weight"]], dim=0
        )
        assert framework.is_equal(actual, exp)
    elif framework.get_name() == "paddle":
        import paddle

        exp = paddle.concat(
            [origs["h.0.attn.c_proj.weight"], origs["h.1.attn.c_proj.weight"]], axis=0
        )
        assert framework.is_equal(actual, exp)

    actual = bufs.get_multi_cols(
        ["h.0.mlp.c_proj.weight", "h.1.mlp.c_proj.weight"], dim=1
    )
    if framework.get_name() == "pytorch":
        exp = torch.concat(
            [origs["h.0.mlp.c_proj.weight"], origs["h.1.mlp.c_proj.weight"]], dim=1
        )
        assert framework.is_equal(actual, exp)
    elif framework.get_name() == "paddle":
        exp = paddle.concat(
            [origs["h.0.mlp.c_proj.weight"], origs["h.1.mlp.c_proj.weight"]], axis=1
        )
        assert framework.is_equal(actual, exp)

    bufs.close()
    loader.close()
    assert framework.get_mem_used() == 0
    assert fstcpp.get_cpp_metrics().bounce_buffer_bytes == 0


if __name__ == "__main__":
    import os
    import sys

    os.environ["PADDLE_DISTRI_BACKEND"] = "gloo"
    sys.exit(pytest.main(sys.argv[1:]))



# SPDX-License-Identifier: Apache-2.0

try:
    import paddle
    import paddle.distributed as pdist
    from paddle.distributed.communication.group import Group
    from paddle.framework import core as paddle_core
except ImportError as e:
    raise ImportError(
        "could not import paddle, paddle_core, or numpy. Please install them."
    ) from e

from dataclasses import dataclass
from typing import Any, Dict, List, Optional

from ..common import SingleGroup
from ..cpp import cpu_free, cpu_malloc, gds_device_buffer, gpu_free, gpu_malloc
from ..st_types import Device, DeviceType, DType
from . import FrameworkOpBase, ProcessGroupBase, TensorBase

dtype_convert: Dict[DType, Any] = {
    DType.BOOL: paddle.bool,
    DType.U8: paddle.uint8,
    DType.I8: paddle.int8,
    DType.I16: paddle.int16,
    DType.U16: paddle.bfloat16,
    DType.I32: paddle.int32,
    DType.U32: paddle.int32,
    DType.I64: paddle.int64,
    DType.U64: paddle.int64,
    DType.F16: paddle.float16,
    DType.BF16: paddle.bfloat16,
    DType.F32: paddle.float32,
    DType.F64: paddle.float64,
}
need_workaround_dtypes: Dict[DType, DType] = {
    DType.F8_E5M2: DType.I8,
    DType.F8_E4M3: DType.I8,
    DType.U16: DType.BF16,
}

if hasattr(paddle, "float8_e5m2"):
    dtype_convert[DType.F8_E5M2] = paddle.float8_e5m2
if hasattr(paddle, "float8_e4m3fn"):
    dtype_convert[DType.F8_E4M3] = paddle.float8_e4m3fn


@dataclass
class PaddleTensor(TensorBase):
    real_tensor: paddle.Tensor

    def get_raw(self) -> paddle.Tensor:
        return self.real_tensor

    def contiguous(self) -> "PaddleTensor":
        return PaddleTensor(self.device, self.dtype, self.real_tensor.contiguous())

    def to(
        self,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> "PaddleTensor":
        to_dev: Optional[str] = None
        if device is not None and self.device != device:
            to_dev = device.as_str()
        else:
            device = self.device
        to_dtype: Optional[paddle.dtype] = None
        if dtype != DType.AUTO and (dtype != self.dtype):
            to_dtype = dtype_convert[dtype]
        else:
            dtype = self.dtype
        if to_dev is not None or to_dtype is not None:
            return PaddleTensor(
                device, dtype, self.real_tensor.to(device=to_dev, dtype=to_dtype)
            )
        return self

    def clone(self) -> "PaddleTensor":
        return PaddleTensor(self.device, self.dtype, self.real_tensor.clone())

    def detach(self) -> "PaddleTensor":
        return PaddleTensor(self.device, self.dtype, self.real_tensor.detach())

    def view(self, dtype: DType) -> "PaddleTensor":
        t2 = self.real_tensor.view(dtype_convert[dtype])
        return PaddleTensor(self.device, dtype, t2)

    def __getitem__(self, _val) -> "PaddleTensor":
        return PaddleTensor(self.device, self.dtype, self.real_tensor[_val])


@dataclass
class PaddleProcessGroup(ProcessGroupBase[PaddleTensor]):
    real_pg: Optional[Group]

    def size(self) -> int:
        return self.real_pg.process_group.size() if self.real_pg else 1

    def rank(self) -> int:
        return self.real_pg.process_group.rank() if self.real_pg else 0

    def broadcast(self, dst: PaddleTensor, rank: int) -> None:
        if self.real_pg:
            pdist.broadcast(dst.real_tensor, rank, group=self.real_pg)
            # Synchronize to ensure the NCCL broadcast (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if paddle.device.is_compiled_with_cuda():
                paddle.device.cuda.synchronize()

    def scatter(
        self,
        dst: PaddleTensor,
        scatter_list: List[PaddleTensor],
        src: int,
    ) -> None:
        if self.real_pg:
            sl = [t.real_tensor for t in scatter_list]
            pdist.scatter(
                dst.real_tensor,
                tensor_list=sl,
                src=src,
                group=self.real_pg,
            )
            # Synchronize to ensure the NCCL scatter (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if paddle.device.is_compiled_with_cuda():
                paddle.device.cuda.synchronize()

    def send(
        self,
        t: PaddleTensor,
        dst_rank: int,
        tag: int,
    ) -> None:
        if self.real_pg:
            pdist.send(t.real_tensor, dst_rank, group=self.real_pg)
            # Synchronize to ensure the NCCL recv (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if paddle.device.is_compiled_with_cuda():
                paddle.device.cuda.synchronize()

    def recv(
        self,
        t: PaddleTensor,
        src_rank: int,
        tag: int,
    ) -> None:
        if self.real_pg:
            pdist.recv(t.real_tensor, src_rank, group=self.real_pg)
            # Synchronize to ensure the NCCL recv (which runs on
            # the NCCL internal stream) is fully visible on the default
            # compute stream before callers read the tensor data.
            if paddle.device.is_compiled_with_cuda():
                paddle.device.cuda.synchronize()


class PaddleOp(FrameworkOpBase[PaddleTensor, PaddleProcessGroup]):
    def __init__(self) -> None:
        self.mem_used = 0

    def get_name(self) -> str:
        return "paddle"

    def get_device(self, device: str, pg: PaddleProcessGroup) -> Device:
        dev_index: Optional[int] = None
        try:
            dev_split = device.split(":", 1)
            dev_type = DeviceType(dev_split[0].lower())
            if dev_type != DeviceType.CPU:
                dev_index = 0
                if len(dev_split) > 1:
                    dev_index = int(dev_split[1])
        except ValueError:
            raise ValueError(f"Unknown device: {device}")

        if paddle.device.cuda.device_count() > 0 and pg.real_pg is not None:
            # For single (gpu:x, gpu)
            # gpu:x, like gpu:0, gpu:1, ...
            # For distributed
            # The gpu determines the current rank
            # rank0 use gpu:0, rank1 use gpu:1
            dev_index = pg.rank() % paddle.device.cuda.device_count()
        return Device(dev_type, dev_index)

    def set_device(self, device: Device) -> None:
        if device.type != DeviceType.CPU:
            paddle.set_device(device.as_str())

    def alloc_tensor_memory(self, length: int, dev: Device) -> gds_device_buffer:
        if dev.type == DeviceType.GPU:
            rbuf = gpu_malloc(length)
        else:
            rbuf = cpu_malloc(length)
        self.mem_used += length
        return gds_device_buffer(rbuf, length, dev.type == DeviceType.GPU)

    def free_tensor_memory(self, gbuf: gds_device_buffer, dev: Device) -> None:
        length = gbuf.get_length()
        if dev.type == DeviceType.GPU:
            gpu_free(gbuf.get_base_address())
        else:
            cpu_free(gbuf.get_base_address())
        self.mem_used -= length

    def get_empty_tensor(
        self, shape: List[int], dtype: DType, device: Device
    ) -> PaddleTensor:
        dst = paddle.to_tensor(
            paddle.empty(shape=shape, dtype=dtype_convert[dtype]),
            place=device.as_str(),
        )
        return PaddleTensor(device, dtype, dst)

    def concat_tensors(self, tensors: List[PaddleTensor], dim) -> PaddleTensor:
        ts = [tensor.real_tensor for tensor in tensors]
        return PaddleTensor(
            tensors[0].device, tensors[0].dtype, paddle.concat(ts, axis=dim)
        )

    def get_dtype_size(self, dtype: DType) -> int:
        return paddle_core.size_of_dtype(dtype_convert[dtype])

    def from_dlpack(self, dl_tensor: Any, device: Device, dtype: DType) -> PaddleTensor:
        return PaddleTensor(device, dtype, paddle.utils.dlpack.from_dlpack(dl_tensor))

    def copy_tensor(self, dst: PaddleTensor, src: PaddleTensor) -> None:
        paddle.assign(src.real_tensor, output=dst.real_tensor)
        dst.dtype = src.dtype
        dst.device = src.device

    def get_cuda_ver(self) -> str:
        """Get GPU runtime version with platform indicator.

        Returns a string like 'hip-5.7.0' for ROCm or 'cuda-12.1' for CUDA,
        or '0.0' if no GPU is available. This allows code to distinguish
        between different GPU platforms without using paddle directly.
        """
        if paddle.device.is_compiled_with_cuda():
            # Check if this is ROCm/HIP build
            if paddle.device.is_compiled_with_rocm():
                return f"hip-{paddle.version.cuda()}"
            return f"cuda-{paddle.version.cuda()}"
        return "0.0"

    def get_device_ptr_align(self) -> int:
        CUDA_PTR_ALIGN: int = 16
        return CUDA_PTR_ALIGN

    def as_workaround_dtype(self, dtype: DType) -> DType:
        if dtype in need_workaround_dtypes:
            return need_workaround_dtypes[dtype]
        return dtype

    def get_process_group(self, pg: Optional[Any]) -> PaddleProcessGroup:
        if pg is not None:
            if isinstance(pg, SingleGroup):
                pg = None
            elif not isinstance(pg, Group):
                raise Exception(
                    "pg must be an instance of paddle.distributed.communication.group.Group"
                )
        return PaddleProcessGroup(pg)

    # for testing
    def is_equal(self, wrapped: PaddleTensor, real: Any) -> bool:
        if isinstance(real, paddle.Tensor):
            return paddle.all(wrapped.real_tensor.equal(real))
        raise Exception("real is not paddle.Tensor")

    def randn(self, s: tuple, device: Device, dtype: DType) -> PaddleTensor:
        return PaddleTensor(
            device,
            dtype,
            paddle.randn(s, dtype=dtype_convert[dtype]).to(device=device.as_str()),
        )

    def support_fp8(self) -> bool:
        return DType.F8_E5M2 in dtype_convert

    def get_mem_used(self) -> int:
        return self.mem_used


_op: Optional[PaddleOp] = None


def get_framework_op() -> FrameworkOpBase:
    global _op
    if _op is None:
        _op = PaddleOp()
    return _op



# SPDX-License-Identifier: Apache-2.0

from typing import Dict, List, Optional, Tuple

from . import cpp as fstcpp
from .common import SafeTensorsMetadata, init_logger, is_debug
from .copier.base import CopierInterface, DummyDeviceBuffer
from .frameworks import FrameworkOpBase, ProcessGroupBase, TensorBase
from .st_types import Device, DType

logger = init_logger(__name__)


class LazyTensorFactory:
    def __init__(
        self,
        metadata: SafeTensorsMetadata,
        device: Device,
        rank: int,
        local_rank: bool,
        factory_idx_bits: int,
        lidx: int,
        copier: CopierInterface,
        framework: FrameworkOpBase,
        disable_cache=True,
    ):
        self.framework = framework
        self.metadata = metadata
        self.device = device
        self.copier: Optional[CopierInterface] = None
        if local_rank:
            self.copier = copier
        self.tensors: Dict[str, TensorBase] = {}
        self.shuffled: Dict[str, TensorBase] = {}
        self.gbuf: Optional[fstcpp.gds_device_buffer] = None
        self.rank = rank
        self.factory_idx_bits = factory_idx_bits
        self.lidx = lidx
        self.next_tag = 1
        self.disable_cache = disable_cache

    def submit_io(self, use_buf_register: bool, max_copy_block_size: int):
        if self.copier is not None:
            self.gbuf = self.copier.submit_io(use_buf_register, max_copy_block_size)
            if self.gbuf:
                logger.debug(
                    "submit_io: new buf, addr=0x%x", self.gbuf.get_base_address()
                )

    def wait_io(self, dtype: DType = DType.AUTO, noalign: bool = False):
        if self.copier is not None and self.gbuf is not None:
            self.tensors = self.copier.wait_io(self.gbuf, dtype=dtype, noalign=noalign)
            if is_debug(logger):
                for name in self.tensors.keys():
                    logger.debug("wait_io: tensor=%s", name)
            self.copier = None

    def push(
        self,
        pg: ProcessGroupBase,
        tensor_name: str,
        dst_rank: int,
        src_rank: int,
    ) -> Optional[TensorBase]:
        if pg.size() == 1:
            return self.tensors[tensor_name]
        tag = (self.next_tag << self.factory_idx_bits) + self.lidx
        self.next_tag += 1
        if pg.rank() != dst_rank and pg.rank() != src_rank:
            logger.debug(
                "push: skip, tensor_name=%s, dst_rank=%d, pg.rank()=%d, tag=%d",
                tensor_name,
                dst_rank,
                pg.rank(),
                tag,
            )
            return None
        elif pg.rank() == dst_rank and src_rank == dst_rank:
            logger.debug(
                "push: nocopy, tensor_name=%s, dst_rank=%d, pg.rank()=%d, tag=%d",
                tensor_name,
                dst_rank,
                pg.rank(),
                tag,
            )
            return self.tensors[tensor_name].clone().detach()
        frame = self.metadata.tensors[tensor_name]
        if pg.rank() == src_rank:
            if tensor_name not in self.tensors:
                raise Exception(
                    f"push: tensor {tensor_name} was not found. released? lidx={self.lidx}"
                )
            t = self.tensors[tensor_name].clone().detach()
            logger.debug(
                "push: send, tensor_name=%s, shape=%s, dst_rank=%d, pg.rank()=%d, tag=%d",
                tensor_name,
                frame.shape,
                dst_rank,
                pg.rank(),
                tag,
            )
            pg.send(t, dst_rank, tag=tag)
            return None

        logger.debug(
            "push: recv, tensor_name=%s, shape=%s, src_rank=%d, pg.rank()=%d, tag=%d",
            tensor_name,
            frame.shape,
            src_rank,
            pg.rank(),
            tag,
        )

        t = self.framework.get_empty_tensor(frame.shape, frame.dtype, self.device)
        pg.recv(t, src_rank, tag=tag)
        return t

    def shuffle(self, pg: ProcessGroupBase, tensor_name: str, dim: int) -> TensorBase:
        if pg.size() == 1:
            return self.tensors[tensor_name]
        if tensor_name in self.shuffled:
            logger.debug("shuffle: use cache, tensor_name=%s", tensor_name)
            t = self.shuffled[tensor_name].clone().detach()
            return t
        frame = self.metadata.tensors[tensor_name]
        if dim == -1:
            if tensor_name in self.tensors:
                dst = self.tensors[tensor_name].clone().detach()
            else:
                dst = self.framework.get_empty_tensor(
                    frame.shape, frame.dtype, self.device
                )
            logger.debug(
                "shuffle: broadcast, tensor_name=%s, shape=%s, self.rank=%d, pg.rank()=%d, has_tensor=%s",
                tensor_name,
                frame.shape,
                self.rank,
                pg.rank(),
                tensor_name in self.tensors,
            )
            pg.broadcast(dst, self.rank)
        else:
            rank_slices: List[Tuple] = [() for i in range(0, pg.size())]
            size = frame.shape[dim]
            block_size = (size + pg.size() - 1) // pg.size()
            for rank in range(0, pg.size()):
                for i in range(0, len(frame.shape)):
                    if i < dim:
                        rank_slices[rank] += (slice(None, None, None),)
                    elif i == dim:
                        rank_slices[rank] += (
                            slice(rank * block_size, (rank + 1) * block_size, 1),
                        )
                        break
            scatter_list: List[TensorBase] = []
            new_frame = frame[rank_slices[pg.rank()]]
            dst = self.framework.get_empty_tensor(
                new_frame.shape, new_frame.dtype, self.device
            )
            if self.rank == pg.rank():
                if tensor_name not in self.tensors:
                    raise Exception(
                        f"shuffle: tensor {tensor_name} was not found, released? lidx={self.lidx}"
                    )
                t = self.tensors[tensor_name]
                scatter_list = [
                    t[rank_slices[rank]].contiguous() for rank in range(0, pg.size())
                ]  # scatter requires contiguous tensor
            logger.debug(
                "shuffle: scatter, tensor_name=%s, shape=%s->%s, self.rank=%d, pg.rank()=%d, rank_slices=%s, len(scatter_list)=%s",
                tensor_name,
                frame.shape,
                new_frame.shape,
                self.rank,
                pg.rank(),
                rank_slices,
                len(scatter_list),
            )
            pg.scatter(dst, scatter_list=scatter_list, src=self.rank)
        if not self.disable_cache:
            # Cache tensor for reuse within the same batch to improve performance.
            # Note: This requires additional (GPU) memory to store the cached tensors.
            # Enable this only if you have sufficient (GPU) memory and required.
            self.shuffled[tensor_name] = dst
        return dst

    def shuffle_multi_cols(
        self, pg: ProcessGroupBase, tensor_names: List[str], dim: int
    ) -> TensorBase:
        rank_tensors: List[List[TensorBase]] = [[] for i in range(0, pg.size())]
        new_shape: List[int] = []
        for tensor_name in tensor_names:
            frame = self.metadata.tensors[tensor_name]
            total_size = frame.shape[dim]
            block_size = (total_size + pg.size() - 1) // pg.size()
            if len(new_shape) == 0:
                new_shape = frame.shape
                new_shape[dim] = 0
            else:
                for dim2 in range(0, len(frame.shape)):
                    if dim2 != dim and frame.shape[dim2] != new_shape[dim2]:
                        raise Exception(
                            f"dim {dim2} mismatch: tensor {tensor_name} has {frame.shape} vs. {new_shape} (dim={dim})"
                        )
            new_shape[dim] += block_size
            if self.rank == pg.rank():
                if tensor_name not in self.tensors:
                    raise Exception(
                        f"shuffle_multi_cols: tensor {tensor_name} was not found, released? lidx={self.lidx}"
                    )
                t = self.tensors[tensor_name]
                for rank in range(0, pg.size()):
                    rank_slices: Tuple[slice, ...] = ()
                    for i in range(0, len(frame.shape)):
                        if i < dim:
                            rank_slices += (slice(None, None, None),)
                        elif i == dim:
                            rank_slices += (
                                slice(rank * block_size, (rank + 1) * block_size, 1),
                            )
                            break
                    rank_tensors[rank].append(t[rank_slices])
        if pg.size() == 1:
            return self.framework.concat_tensors(rank_tensors[self.rank], dim=dim)
        scatter_list: List[TensorBase] = []

        if self.rank == pg.rank():
            for rank in range(0, pg.size()):
                scatter_list.append(
                    self.framework.concat_tensors(rank_tensors[rank], dim=dim)
                )
        logger.debug(
            "shuffle_multi_cols: scatter, tensor_name=%s, shape=%s->%s, self.rank=%d, pg.rank()=%d, len(scatter_list)=%s",
            tensor_name,
            frame.shape,
            new_shape,
            self.rank,
            pg.rank(),
            len(scatter_list),
        )
        dst = self.framework.get_empty_tensor(new_shape, frame.dtype, self.device)
        pg.scatter(dst, scatter_list=scatter_list, src=self.rank)
        return dst

    def free_dev_ptrs(self):
        self.tensors = {}
        if self.gbuf is not None and not isinstance(self.gbuf, DummyDeviceBuffer):
            self.framework.free_tensor_memory(self.gbuf, self.device)
            logger.debug(
                "free_dev_ptrs: delete buf, addr=0x%x", self.gbuf.get_base_address()
            )
            self.gbuf = None



# SPDX-License-Identifier: Apache-2.0

import os
import queue
import threading
import time
from typing import Any, Generator, List, Optional, Tuple, Union

import torch
from tqdm.auto import tqdm

from . import BaseSafeTensorsFileLoader, SafeTensorsFileLoader
from . import cpp as fstcpp


def enable_tqdm(use_tqdm_on_load: bool):
    """Determine whether to enable tqdm progress bar based on distributed settings.

    Progress bar is enabled only on rank 0 when in distributed mode, or always
    enabled in single GPU mode.

    Args:
        use_tqdm_on_load: User preference for enabling tqdm

    Returns:
        bool: True if tqdm should be enabled, False otherwise
    """
    return use_tqdm_on_load and (
        not torch.distributed.is_initialized() or torch.distributed.get_rank() == 0
    )


_BAR_FORMAT = "{desc}: {percentage:3.0f}% Completed | {n_fmt}/{total_fmt} [{elapsed}<{remaining}, {rate_fmt}]\n"


def parse_time(timestamp_ms):
    """Convert timestamp in milliseconds to formatted string with millisecond precision.

    Args:
        timestamp_ms: Timestamp in milliseconds

    Returns:
        str: Formatted timestamp string with millisecond precision
    """
    return (
        time.strftime("%Y-%m-%d %H:%M:%S", time.localtime(timestamp_ms))
        + f".{int((timestamp_ms % 1) * 1000):03d}"
    )


class TimingContext:
    """Context manager for timing operations."""

    def __init__(
        self,
        name: str,
        logger=None,
        batch_id=None,
        threshold_ms: float = 0,
        log_on_exit: bool = True,
    ):
        self.name = name
        self.logger = logger
        self.threshold_ms = threshold_ms
        self.log_on_exit = log_on_exit
        self.start_time = None
        self.elapsed_ms = 0
        self.batch_id = batch_id

    def __enter__(self):
        self.start_time = time.time()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        msg = ""
        if self.batch_id is not None:
            msg += f"Batch {self.batch_id}: "
        self.elapsed_ms = (time.time() - self.start_time) * 1000
        if (
            self.logger
            and self.log_on_exit
            and (self.elapsed_ms >= self.threshold_ms or exc_type is not None)
        ):
            msg += f"{self.name} took {self.elapsed_ms:.3f} ms"
            self.logger(msg)
        return False


class FileBatch:
    """Represents a batch of files loaded into device memory.

    This class encapsulates a batch of SafeTensors files that have been loaded
    into memory, along with metadata about the loading process for performance
    monitoring and debugging purposes.
    """

    def __init__(self, fb, keys: List[str], batch_id: int):
        """Initialize a FileBatch instance.

        Args:
            fb: SafeTensors file buffer containing loaded tensors
            keys: List of tensor keys available in this batch
            batch_id: Unique identifier for this batch
        """
        self.fb = fb  # SafeTensors file buffer
        self.keys = keys  # tensor keys
        self.batch_id = batch_id
        self.load_time: Optional[float] = None  # Record loading time for debugging
        self.add_filenames_time: Optional[float] = None  # Record add_filenames time
        self.copy_files_time: Optional[float] = None  # Record copy_files_to_device time


class PipelineParallel:
    r"""Load .safetensors files in parallel with producer-consumer pattern."""

    def __init__(
        self,
        pg: Optional[Any],
        loader: BaseSafeTensorsFileLoader,
        hf_weights_files: List[str],
        max_concurrent_producers: int = 1,
        queue_size: int = 0,  # Changed default to 0 for unbuffered behavior
        use_tqdm_on_load: bool = True,
        **kwargs,
    ):

        self.loader = loader
        self.hf_weights_files = hf_weights_files
        self.max_concurrent_producers = max_concurrent_producers
        self.queue_size = queue_size
        self.use_tqdm_on_load = use_tqdm_on_load

        # Batch files
        self.weight_files_batches = self._create_batches(pg)

        # Producer-consumer communication
        # For unbuffered behavior (queue_size=0), we use a maxsize of 1 to ensure synchronization
        # but modify the producer logic to wait for consumer to process before producing next
        self.batch_queue: queue.Queue[Union[FileBatch, Exception, None]] = queue.Queue(
            maxsize=max(1, queue_size)
        )  # Ensure at least size 1
        self.stop_event = threading.Event()
        self.error_event = threading.Event()
        self.error_info: Optional[str] = None

        # For unbuffered behavior, we need additional synchronization
        self.consumer_processed: Optional[threading.Event] = (
            threading.Event() if queue_size <= 0 else None
        )
        if queue_size <= 0 and self.consumer_processed is not None:
            self.consumer_processed.set()  # Initially set to allow first production

        # Logging setup - get from environment variable, default to False
        self.print_log = os.getenv("FASTSAFETENSORS_DEBUG", "false").lower() == "true"
        self.log_prefix = f"PG{pg.rank() if pg is not None else 0}"
        fstcpp.set_gil_release(True)

    def _create_batches(self, pg) -> List[List[str]]:
        """Create file batches based on distributed settings.

        In distributed mode, files are grouped by the process group size so that
        each process in the group handles one file from each batch.

        Args:
            pg: Process group for distributed operations

        Returns:
            List[List[str]]: List of file batches, where each batch contains
                            files for all processes in the group
        """
        batch_size = pg.size()
        return [
            self.hf_weights_files[i : i + batch_size]
            for i in range(0, len(self.hf_weights_files), batch_size)
        ]

    def _log_message(self, message: str, is_error: bool = False):
        """Unified logging method for conditional print statements.
        Args:
            message: The message to log
        """
        t = parse_time(time.time())
        if self.print_log or is_error:
            print(f"[{self.log_prefix}] {t} - {message}", flush=True)

    def _log_error(self, message: str):
        self._log_message(message, is_error=True)

    def _load_single_batch(self, batch_id: int, file_list: List[str]):
        """Load a single batch into device memory.

        This method handles the complete process of loading a batch of files:
        1. Preparing file mappings
        2. Adding filenames to the loader
        3. Copying files to device memory
        4. Creating a FileBatch object with timing information
        5. Putting the batch into the queue for consumer processing

        Args:
            batch_id: Identifier for this batch
            file_list: List of files to load in this batch
        """
        if self.stop_event.is_set() or self.error_event.is_set():
            self._log_message(
                f"Producer batch {batch_id} skipped: stop_event-{self.stop_event.is_set()} "
                f"or error_event-{self.error_event.is_set()},"
                f"error_info-{self.error_info}",
                is_error=True,
            )
            return

        try:
            # Prepare file mapping
            rank_file_map = {i: [f] for i, f in enumerate(file_list)}

            with TimingContext("add_filenames", self._log_message, batch_id) as timer:
                self.loader.add_filenames(rank_file_map)
            add_filenames_time = timer.elapsed_ms

            # For unbuffered behavior, wait for consumer to process previous item
            if self.queue_size <= 0 and self.consumer_processed is not None:
                if not self.consumer_processed.wait():
                    raise TimeoutError(
                        "can not waiting for consumer to process previous batch"
                    )
                # Clear the event after wait to ensure next wait will block
                self.consumer_processed.clear()

            with TimingContext(
                "copy_files_to_device", self._log_message, batch_id
            ) as timer:
                fb = self.loader.copy_files_to_device()
            copy_time = timer.elapsed_ms

            # Get tensor keys
            keys = list(fb.key_to_rank_lidx.keys())

            # Create batch object with timing info
            batch = FileBatch(fb, keys, batch_id)
            batch.add_filenames_time = add_filenames_time
            batch.copy_files_time = copy_time
            batch.load_time = add_filenames_time + copy_time  # Total load time

            # Put into queue for consumer processing
            if not self.stop_event.is_set():
                self.batch_queue.put(batch)
                with TimingContext("loader.reset", self._log_message, batch_id):
                    self.loader.reset()

        except Exception as e:
            self.error_info = f"Producer batch {batch_id} failed: {e}"
            self.error_event.set()
            self.batch_queue.put(e)  # Notify consumer of error

    def _producer_worker(self):
        """Producer worker thread: responsible for copy_files_to_device operations.

        This method runs in a separate thread and handles the production of file
        batches by loading them into memory and placing them in the queue for
        consumption by the main thread.
        """
        batch_id = 0
        try:
            for f_list in tqdm(
                self.weight_files_batches,
                desc="Loading fastsafetensors checkpoint shards",
                disable=True,
                bar_format=_BAR_FORMAT,
            ):
                self._load_single_batch(batch_id, f_list)
                batch_id += 1
        except Exception as e:
            if not self.error_event.is_set():
                self.error_info = f"Producer future failed: {e}"
                self.error_event.set()
                self.batch_queue.put(e)

        # Signal end of production
        self.batch_queue.put(None)

    def _consume_single_batch(self):
        with TimingContext("wait_queue", self._log_message) as timer:
            batch_item = self.batch_queue.get()
            # For unbuffered behavior, signal that we've processed this item
            if self.queue_size == 0 and self.consumer_processed is not None:
                self.consumer_processed.set()
            # Check end signal
            if batch_item is None:
                self._log_error(f"get batch_item is None, will break")
                return
            # Check for errors
            if isinstance(batch_item, Exception):
                self._log_error(f"get batch_item is Exception, will raise")
                raise batch_item
            # Process normal batch
            batch = batch_item
            timer.batch_id = batch.batch_id
        queue_wait_time = timer.elapsed_ms
        if queue_wait_time / 1000 > 10:  # Only log if wait time is significant > 10s
            self._log_message(
                f"Warning: Batch {batch.batch_id}: Queue wait took {queue_wait_time:.3f} ms",
                is_error=True,
            )
        try:
            self._log_message(
                f"Batch {batch.batch_id}: tensor key len: {len(batch.keys)}"
            )
            # Consumer operation: extract tensors
            with TimingContext(
                "get_tensor", self._log_message, batch.batch_id
            ) as timer:
                for key in batch.keys:
                    tensor = batch.fb.get_tensor(key)
                    yield key, tensor
            get_tensor_time = timer.elapsed_ms
        finally:
            # Close the file buffer
            with TimingContext("fb.close", self._log_message, batch.batch_id) as timer:
                batch.fb.close()
            close_time = timer.elapsed_ms
        # Log batch summary with all timing info
        self._log_message(
            f"Batch {batch.batch_id} summary: "
            f"add_filenames={batch.add_filenames_time:.3f}ms, "
            f"copy_files={batch.copy_files_time:.3f}ms, "
            f"get_tensor={get_tensor_time:.3f}ms, "
            f"close={close_time:.3f}ms"
        )
        # sync
        if self.queue_size < 0 and self.consumer_processed is not None:
            self.consumer_processed.set()

    def _consumer_worker(self):
        start_time_total = time.time()
        processed_batches = 0
        try:
            for _ in tqdm(
                self.weight_files_batches,
                desc="Loading fastsafetensors checkpoint shards",
                disable=not enable_tqdm(self.use_tqdm_on_load),
                bar_format=_BAR_FORMAT,
            ):
                yield from self._consume_single_batch()
                processed_batches += 1
        except Exception as e:
            self._log_message(f"Consumer error: {e}", is_error=True)
            self.stop_event.set()
            raise
        finally:
            end_time_total = time.time()
            elapsed_time = end_time_total - start_time_total
            self._log_message(
                f"Completed ParallelLoader iterate_weights, "
                f"processed {processed_batches} batches, total time: {elapsed_time:.2f} seconds"
            )
        if processed_batches < len(self.weight_files_batches):
            self._log_error(f"Unexpected Error: not all tensors has been exported")

    def iterate_weights(self) -> Generator[Tuple[str, torch.Tensor], None, None]:
        """Main weight iterator: consumer logic.

        This method implements the consumer side of the producer-consumer pattern.
        It retrieves batches from the queue, extracts tensors, and yields them
        one by one. It also handles cleanup and error reporting.

        Yields:
            Tuple[str, torch.Tensor]: Key-value pairs of tensor names and tensors
        """
        self._log_message("Starting ParallelLoader iterate_weights")

        # Start producer thread
        producer_thread = threading.Thread(target=self._producer_worker)
        producer_thread.start()

        try:
            yield from self._consumer_worker()
        finally:
            # Cleanup work
            self.stop_event.set()
            producer_thread.join(timeout=5)

    def close(self):
        self.loader.close()


class ParallelLoader(PipelineParallel):
    r"""Load .safetensors files in parallel with producer-consumer pattern.

    This class is a convenience wrapper around PipelineParallel that handles
    the creation of the SafeTensorsFileLoader with appropriate parameters.

    Args:
        pg (Optional[Any]): Process group-like objects for distributed operations.
                           None for single GPU use-cases.
        hf_weights_files (List[str]): List of safetensors files to load.
        max_concurrent_producers (int): Maximum number of concurrent producer threads
                                       for file loading. Currently only 1 is supported.
        queue_size (int): Size of the queue for buffering loaded file batches.
                         Set to 0 for unbuffered behavior.
        use_tqdm_on_load (bool): Enable progress bar during loading.
        device (str): Target device for tensor loading (e.g., "cpu", "cuda:0").
        bbuf_size_kb (int): Bounce buffer size for file copies in KB.
        max_threads (int): Maximum number of threads for memory copies.
        nogds (bool): If True, turn off GDS and fallback to pread with bounce buffer.
        use_shm (bool): If True, use shared memory for file loading.
        set_numa (bool): If True, set NUMA node for optimal memory allocation.
        debug_log (bool): Enable debug logs.
        framework (str): Framework to use for tensor operations, e.g., "pytorch".

    Additional GPU memory consumption: (max_concurrent_producers + queue_size) * file_size
    To reduce GPU memory consumption, re-accessing tensors that have already been accessed is prohibited.

    Examples:
        >> from fastsafetensors import ParallelLoader
        >> src_files = download(target_dir, "gpt2")
        >> iterator = ParallelLoader(None, src_files, max_concurrent_producers=1, queue_size=0)
        >> for key, tensor in iterator.iterate_weights():
        >>     print(f"Loaded tensor: {key}, shape: {tensor.shape}")
    """

    def __init__(
        self,
        pg: Optional[Any],
        hf_weights_files: List[str],
        max_concurrent_producers: int = 1,
        queue_size: int = 0,  # Changed default to 0 for unbuffered behavior
        use_tqdm_on_load: bool = True,
        device: str = "cpu",
        bbuf_size_kb: int = 16 * 1024,
        max_threads: int = 16,
        nogds: bool = False,
        set_numa: bool = True,
        debug_log: bool = False,
        framework="pytorch",
        **kwargs,
    ):
        """Initialize PipelineParallelLoader with a pre-configured SafeTensorsFileLoader.

        Args:
            pg (Optional[Any]): Process group-like objects for distributed operations.
            hf_weights_files (List[str]): List of safetensors files to load.
            max_concurrent_producers (int): Maximum number of concurrent producer threads.
            queue_size (int): Size of the queue for buffering loaded file batches.
            use_tqdm_on_load (bool): Enable progress bar during loading.
            device (str): Target device for tensor loading.
            bbuf_size_kb (int): Bounce buffer size for file copies in KB.
            max_threads (int): Maximum number of threads for memory copies.
            nogds (bool): If True, turn off GDS and fallback to pread with bounce buffer.
            set_numa (bool): If True, set NUMA node for optimal memory allocation.
            debug_log (bool): Enable debug logs.
            framework (str): Framework to use for tensor operations
        """
        loader = SafeTensorsFileLoader(
            pg,
            device,
            bbuf_size_kb=bbuf_size_kb,
            max_threads=max_threads,
            nogds=nogds,
            disable_cache=True,
            set_numa=set_numa,
            debug_log=debug_log,
            framework=framework,
            **kwargs,
        )
        super().__init__(
            pg,
            loader,
            hf_weights_files,
            max_concurrent_producers,
            queue_size,
            use_tqdm_on_load,
            **kwargs,
        )



# SPDX-License-Identifier: Apache-2.0

class gds_device_buffer:
    def __init__(self, devPtr_base: int, length: int, use_cuda: bool) -> None: ...
    def cufile_register(self, offset: int, length: int) -> int: ...
    def cufile_deregister(self, offset: int) -> int: ...
    def memmove(
        self, dst_off: int, src_off: int, tmp: "gds_device_buffer", length: int
    ) -> int: ...
    def get_base_address(self) -> int: ...
    def get_length(self) -> int: ...

class nogds_file_reader:
    def __init__(
        self,
        use_mmap: bool,
        bbuf_size_kb: int,
        max_threads: int,
        use_cuda: bool,
        device_id: int,
    ) -> None: ...
    def submit_read(
        self, fd: int, dst: gds_device_buffer, offset: int, length: int, ptr_off: int
    ) -> int: ...
    def wait_read(self, thread_id: int) -> int: ...

class gds_file_handle:
    def __init__(self, filename: str, o_direct: bool, use_cuda: bool) -> None: ...

class gds_file_reader:
    def __init__(self, max_threads: int, use_cuda: bool, device_id: int) -> None: ...
    def submit_read(
        self,
        fh: gds_file_handle,
        dst: gds_device_buffer,
        offset: int,
        length: int,
        ptr_off: int,
        file_length: int,
    ) -> int: ...
    def wait_read(self, id: int) -> int: ...

class cpp_metrics:
    bounce_buffer_bytes: int

    def __init__(self) -> None: ...

def is_cuda_found() -> bool: ...
def is_cufile_found() -> bool: ...
def cufile_version() -> int: ...
def get_alignment_size() -> int: ...
def set_debug_log(debug_log: bool) -> None: ...
def is_gds_supported(deviceId: int) -> int: ...
def init_gds() -> int: ...
def close_gds() -> int: ...
def get_device_pci_bus(deviceId: int) -> str: ...
def set_numa_node(numa_node: int) -> int: ...
def read_buffer(dst: int, length: int) -> bytes: ...
def cpu_malloc(length: int) -> int: ...
def cpu_free(addr: int) -> None: ...
def gpu_malloc(length: int) -> int: ...
def gpu_free(addr: int) -> None: ...
def load_library_functions() -> None: ...
def get_cpp_metrics() -> cpp_metrics: ...
def set_gil_release(gil_release: bool) -> None: ...
def get_gil_release() -> bool: ...



# SPDX-License-Identifier: Apache-2.0

from importlib.metadata import version

__version__ = version(__name__)

from .common import (
    SafeTensorsMetadata,
    SingleGroup,
    TensorFrame,
    get_device_numa_node,
)
from .file_buffer import FilesBufferOnDevice
from .loader import BaseSafeTensorsFileLoader, SafeTensorsFileLoader, fastsafe_open
from .parallel_loader import ParallelLoader



# SPDX-License-Identifier: Apache-2.0
# Copyright 2017 by Contributors
# Copyright 2024 IBM Inc. All rights reserved
# modified apps/numpy_dlpack/dlpack/dlpack.py and apps/numpy_dlpack/dlpack/from_numpy.py in https://github.com/dmlc/dlpack
# to add from_cuda_buffer()

import ctypes
from typing import Dict, List, Union

from .st_types import Device, DeviceType, DType

_c_str_dltensor = b"dltensor"


# Lazy GPU type detection - avoid calling framework-specific code at module load time
_GPU_DEVICE_TYPE = None  # Will be detected lazily


def _detect_gpu_type():
    """Detect if we're running on ROCm or CUDA.

    This detection is now done lazily to avoid framework-specific calls at module load time.
    Uses the C++ extension's is_hip_found() to determine the platform.
    """
    # Import here to avoid circular dependency
    from . import cpp as fstcpp

    # Check if we loaded HIP runtime (ROCm)
    if fstcpp.is_hip_found():
        return 10  # kDLROCM
    return 2  # kDLCUDA


def _get_gpu_device_type():
    """Get the GPU device type, detecting it lazily if needed."""
    global _GPU_DEVICE_TYPE
    if _GPU_DEVICE_TYPE is None:
        _GPU_DEVICE_TYPE = _detect_gpu_type()
    return _GPU_DEVICE_TYPE


class DLDevice(ctypes.Structure):
    def __init__(self, dev: Device):
        # Use lazy detection to get the GPU device type
        gpu_type = _get_gpu_device_type()
        device_to_dl = {
            DeviceType.CPU: self.kDLCPU,
            DeviceType.CUDA: gpu_type,
            DeviceType.GPU: gpu_type,
        }
        self.device_type = device_to_dl[dev.type]
        self.device_id = dev.index if dev.index is not None else 0

    kDLCPU = 1
    kDLCUDA = 2
    kDLROCM = 10
    _fields_ = [
        ("device_type", ctypes.c_int),
        ("device_id", ctypes.c_int),
    ]


class c_DLDataType(ctypes.Structure):
    def __init__(self, dtype: DType):
        self.type_code, self.bits, self.lanes = self.STDataToDL[dtype]

    kDLInt = 0
    kDLUInt = 1
    kDLFloat = 2
    kDLBfloat = 4
    kDLBool = 6
    _fields_ = [
        ("type_code", ctypes.c_uint8),
        ("bits", ctypes.c_uint8),
        ("lanes", ctypes.c_uint16),
    ]

    STDataToDL: Dict[DType, tuple[int, int, int]] = {
        DType.BOOL: (kDLBool, 8, 1),
        DType.I8: (kDLInt, 8, 1),
        DType.I16: (kDLInt, 16, 1),
        DType.I32: (kDLInt, 32, 1),
        DType.I64: (kDLInt, 64, 1),
        DType.U8: (kDLUInt, 8, 1),
        DType.U16: (kDLUInt, 16, 1),
        DType.U32: (kDLUInt, 32, 1),
        DType.U64: (kDLUInt, 64, 1),
        DType.F16: (kDLFloat, 16, 1),
        DType.F32: (kDLFloat, 32, 1),
        DType.F64: (kDLFloat, 64, 1),
        DType.BF16: (kDLBfloat, 16, 1),
    }


class _Holder:
    def __init__(self, shape: List[int], strides: List[int]):
        self.shape = (ctypes.c_int64 * len(shape))(*shape)
        self.strides = (ctypes.c_int64 * len(strides))(*strides)

    def _as_manager_ctx(self) -> ctypes.c_void_p:
        py_obj = ctypes.py_object(self)
        py_obj_ptr = ctypes.pointer(py_obj)
        ctypes.pythonapi.Py_IncRef(py_obj)
        ctypes.pythonapi.Py_IncRef(ctypes.py_object(py_obj_ptr))
        return ctypes.cast(py_obj_ptr, ctypes.c_void_p)


class DLTensor(ctypes.Structure):
    _fields_ = [
        ("data", ctypes.c_void_p),
        ("device", DLDevice),
        ("ndim", ctypes.c_int),
        ("dtype", c_DLDataType),
        ("shape", ctypes.POINTER(ctypes.c_int64)),
        ("strides", ctypes.POINTER(ctypes.c_int64)),
        ("byte_offset", ctypes.c_uint64),
    ]

    def __init__(self, dev_ptr: int, dev: Device, dtype: DType, holder: _Holder):
        self.data = dev_ptr
        self.device = DLDevice(dev)
        self.ndim = len(holder.shape)
        self.dtype = c_DLDataType(dtype)
        self.shape = holder.shape
        self.strides = holder.strides
        self.byte_offset = 0

    @property
    def itemsize(self):
        return self.dtype.lanes * self.dtype.bits // 8

    @property
    def __array_interface__(self):
        shape = tuple(self.shape[dim] for dim in range(self.ndim))
        if self.strides:
            strides = tuple(
                self.strides[dim] * self.itemsize for dim in range(self.ndim)
            )
        else:
            # Array is compact, make it numpy compatible.
            strides = []
            for i, s in enumerate(shape):
                cumulative = 1
                for e in range(i + 1, self.ndim):
                    cumulative *= shape[e]
                strides.append(cumulative * self.itemsize)
            strides = tuple(strides)
        typestr = "|" + str(self.dtype.type_code)[0] + str(self.itemsize)
        return dict(
            version=3,
            shape=shape,
            strides=strides,
            data=(self.data, True),
            offset=self.byte_offset,
            typestr=typestr,
        )


class DLManagedTensor(ctypes.Structure):
    _fields_ = [
        ("dl_tensor", DLTensor),
        ("manager_ctx", ctypes.c_void_p),
        ("deleter", ctypes.CFUNCTYPE(None, ctypes.c_void_p)),
    ]

    def as_py(
        self,
        dev_ptr: int,
        shape: List[int],
        strides: List[int],
        dtype: DType,
        dev: Device,
    ):
        holder = _Holder(shape, strides)
        self.dl_tensor = DLTensor(dev_ptr, dev, dtype, holder)
        self.manager_ctx = holder._as_manager_ctx()
        self.deleter = _numpy_buffer_deleter
        return ctypes.pythonapi.PyCapsule_New(
            ctypes.byref(self),
            _c_str_dltensor,
            _numpy_pycapsule_deleter,
        )

    @property
    def __array_interface__(self):
        return self.dl_tensor.__array_interface__


ctypes.pythonapi.PyMem_RawMalloc.restype = ctypes.c_void_p
ctypes.pythonapi.PyMem_RawFree.argtypes = [ctypes.c_void_p]

ctypes.pythonapi.PyCapsule_New.restype = ctypes.py_object
ctypes.pythonapi.PyCapsule_New.argtypes = [
    ctypes.c_void_p,
    ctypes.c_char_p,
    ctypes.c_void_p,
]


@ctypes.CFUNCTYPE(None, ctypes.c_void_p)
def _numpy_buffer_deleter(handle: Union[int, ctypes.c_void_p]) -> None:
    """A function to deallocate the memory of a cuda buffer."""
    if isinstance(handle, int):
        dl_managed_tensor = DLManagedTensor.from_address(handle)
    elif isinstance(handle, ctypes.c_void_p):
        dl_managed_tensor = DLManagedTensor.from_address(
            handle.value if handle.value else 0
        )
    else:
        raise Exception("invalid type of handle!")
    py_obj_ptr = ctypes.cast(
        dl_managed_tensor.manager_ctx, ctypes.POINTER(ctypes.py_object)
    )
    py_obj = py_obj_ptr.contents
    ctypes.pythonapi.Py_DecRef(py_obj)
    ctypes.pythonapi.Py_DecRef(ctypes.py_object(py_obj_ptr))
    ctypes.pythonapi.PyMem_RawFree(handle)


@ctypes.CFUNCTYPE(None, ctypes.c_void_p)
def _numpy_pycapsule_deleter(handle: ctypes.c_void_p) -> None:
    """A function to deallocate a pycapsule that wraps a cuda buffer."""
    pycapsule: ctypes.py_object = ctypes.cast(handle, ctypes.py_object)
    if ctypes.pythonapi.PyCapsule_IsValid(pycapsule, _c_str_dltensor):
        dl_managed_tensor = ctypes.pythonapi.PyCapsule_GetPointer(
            pycapsule, _c_str_dltensor
        )
        _numpy_buffer_deleter(dl_managed_tensor)
        ctypes.pythonapi.PyCapsule_SetDestructor(pycapsule, None)


def from_cuda_buffer(
    dev_ptr: int, shape: List[int], strides: List[int], dtype: DType, dev: Device
):
    size = ctypes.c_size_t(ctypes.sizeof(DLManagedTensor))
    dl_managed_tensor = DLManagedTensor.from_address(
        ctypes.pythonapi.PyMem_RawMalloc(size)
    )
    return dl_managed_tensor.as_py(dev_ptr, shape, strides, dtype, dev)



# SPDX-License-Identifier: Apache-2.0

from abc import ABC, abstractmethod
from typing import Dict

from .. import cpp as fstcpp
from ..frameworks import TensorBase
from ..st_types import DType


class CopierInterface(ABC):
    @abstractmethod
    def submit_io(
        self, use_buf_register: bool, max_copy_block_size: int
    ) -> fstcpp.gds_device_buffer:
        pass

    @abstractmethod
    def wait_io(
        self,
        gbuf: fstcpp.gds_device_buffer,
        dtype: DType = DType.AUTO,
        noalign: bool = False,
    ) -> Dict[str, TensorBase]:
        pass


class DummyDeviceBuffer(fstcpp.gds_device_buffer):
    def __init__(self):
        super().__init__(0, 0, False)



# SPDX-License-Identifier: Apache-2.0

import os
import re
import shutil
import subprocess
from pathlib import Path

from setuptools import Extension, setup
from setuptools.command.build_ext import build_ext


def detect_platform():
    """
    Detect if we're on NVIDIA CUDA or AMD ROCm platform.

    Returns:
        tuple: (platform_type, rocm_version, rocm_path)
            platform_type: 'cuda' or 'rocm'
            rocm_version: ROCm version string (e.g., '7.0.1') or None
            rocm_path: Path to ROCm installation or None
    """
    # Safely get ROCM_PATH from environment variables, or use the default
    rocm_path = os.environ.get("ROCM_PATH", "/opt/rocm")

    # Check if ROCm is available
    if rocm_path and os.path.exists(rocm_path):
        # Detect ROCm version
        rocm_version = None
        version_file = os.path.join(rocm_path, ".info", "version")
        if os.path.exists(version_file):
            with open(version_file, "r") as f:
                rocm_version = f.read().strip()
        else:
            # Try to extract version from path
            match = re.search(r"rocm[-/](\d+\.\d+(?:\.\d+)?)", rocm_path)
            if match:
                rocm_version = match.group(1)

        print(f"Detected ROCm platform at {rocm_path}")
        if rocm_version:
            print(f"ROCm version: {rocm_version}")
        return ("rocm", rocm_version, rocm_path)

    # Check for CUDA
    cuda_home = os.environ.get("CUDA_HOME") or os.environ.get("CUDA_PATH")
    if not cuda_home:
        # Try to find nvcc
        nvcc_path = shutil.which("nvcc")
        if nvcc_path:
            cuda_home = os.path.dirname(os.path.dirname(nvcc_path))

    if cuda_home and os.path.exists(cuda_home):
        print(f"Detected CUDA platform at {cuda_home}")
        return ("cuda", None, None)

    # Default to CUDA if nothing detected
    print("No GPU platform detected, defaulting to CUDA")
    return ("cuda", None, None)


def hipify_source_files(rocm_path):
    """
    Automatically hipify CUDA source files to HIP using hipify-perl from ROCm.
    The cuda_compat.h header handles what hipify doesn't convert.

    Args:
        rocm_path: Path to ROCm installation

    Returns:
        list: Paths to hipified source files
    """
    cpp_dir = Path("fastsafetensors/cpp").resolve()
    hip_dir = cpp_dir / "hip"

    # Create hip/ subdirectory if it doesn't exist
    hip_dir.mkdir(exist_ok=True)

    # Find hipify-perl in ROCm installation
    hipify_perl = os.path.join(rocm_path, "bin", "hipify-perl")
    if not os.path.exists(hipify_perl):
        raise RuntimeError(
            f"hipify-perl not found at {hipify_perl}. "
            f"Please ensure ROCm is properly installed at {rocm_path}"
        )

    # Files to hipify
    source_files = [
        ("ext.cpp", "ext.cpp"),
        ("ext.hpp", "ext.hpp"),
    ]

    print(f"Hipifying files using hipify-perl from {hipify_perl}:")
    hipified_files = []

    for src_name, dst_name in source_files:
        src_path = cpp_dir / src_name
        dst_path = hip_dir / dst_name

        print(f"  - {src_path} -> {dst_path}")

        try:
            # Run hipify-perl: hipify-perl input.cpp -o output.cpp
            result = subprocess.run(
                [hipify_perl, str(src_path), "-o", str(dst_path)],
                check=True,
                capture_output=True,
                text=True,
            )
            print(f"    Successfully hipified: {src_name}")
            hipified_files.append(str(dst_path))

            # Print any warnings from hipify-perl
            if result.stderr:
                print(f"    hipify-perl output: {result.stderr.strip()}")

            # Post-process: Replace cuda_compat.h with hip_compat.h
            # hipify-perl doesn't convert custom header names
            with open(dst_path, "r") as f:
                content = f.read()
            content = content.replace(
                '#include "cuda_compat.h"', '#include "hip_compat.h"'
            )
            with open(dst_path, "w") as f:
                f.write(content)
            print(f"    Post-processed: cuda_compat.h -> hip_compat.h")

        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"Failed to hipify {src_path}:\n"
                f"stdout: {e.stdout}\n"
                f"stderr: {e.stderr}"
            ) from e

    # Copy cuda_compat.h to hip directory as hip_compat.h
    # (hipify converts the include statement from cuda_compat.h to hip_compat.h)
    cuda_compat = cpp_dir / "cuda_compat.h"
    hip_compat = hip_dir / "hip_compat.h"
    shutil.copy2(cuda_compat, hip_compat)
    print(f"Copied {cuda_compat} -> {hip_compat}")

    return hipified_files


def MyExtension(
    name, sources, mod_name, platform_type, rocm_path=None, *args, **kwargs
):
    import pybind11

    pybind11_path = os.path.dirname(pybind11.__file__)

    kwargs["define_macros"] = [("__MOD_NAME__", mod_name)]
    kwargs["libraries"] = ["stdc++"]
    kwargs["include_dirs"] = kwargs.get("include_dirs", []) + [
        f"{pybind11_path}/include"
    ]  # for pybind11/pybind11.h
    kwargs["language"] = "c++"

    # https://pybind11.readthedocs.io/en/stable/faq.html#someclass-declared-with-greater-visibility-than-the-type-of-its-field-someclass-member-wattributes
    kwargs["extra_compile_args"] = ["-fvisibility=hidden", "-std=c++17"]

    # Platform-specific configuration
    if platform_type == "rocm" and rocm_path:
        # ROCm/HIP configuration
        kwargs["define_macros"].append(("__HIP_PLATFORM_AMD__", "1"))
        kwargs["libraries"].append("amdhip64")
        kwargs["library_dirs"] = [f"{rocm_path}/lib"]
        kwargs["include_dirs"].append(f"{rocm_path}/include")
        kwargs["extra_compile_args"].append("-D__HIP_PLATFORM_AMD__")
        kwargs["extra_link_args"] = [f"-L{rocm_path}/lib", "-lamdhip64"]

    return Extension(name, sources, *args, **kwargs)


class CustomBuildExt(build_ext):
    """Custom build_ext to handle automatic hipification for ROCm platforms"""

    def run(self):
        # Detect platform
        platform_type, rocm_version, rocm_path = detect_platform()

        # Store platform info
        self.platform_type = platform_type
        self.rocm_version = rocm_version
        self.rocm_path = rocm_path

        #  Configure build based on platform
        if platform_type == "rocm" and rocm_path:
            print("=" * 60)
            print("Building for AMD ROCm platform")
            if rocm_version:
                print(f"ROCm version: {rocm_version}")
            print("=" * 60)

            # Hipify sources
            hipify_source_files(rocm_path)

            # Update extension sources to use hipified files
            for ext in self.extensions:
                new_sources = []
                for src in ext.sources:
                    if "fastsafetensors/cpp/ext.cpp" in src:
                        # torch.utils.hipify creates files in hip/ subdirectory
                        new_sources.append(
                            src.replace(
                                "fastsafetensors/cpp/ext.cpp",
                                "fastsafetensors/cpp/hip/ext.cpp",
                            )
                        )
                    else:
                        new_sources.append(src)
                ext.sources = new_sources

                # Update include dirs to include hip/ subdirectory
                ext.include_dirs.append("fastsafetensors/cpp/hip")

                # Update extension with ROCm-specific settings
                ext.define_macros.append(("__HIP_PLATFORM_AMD__", "1"))
                ext.define_macros.append(("USE_ROCM", "1"))
                ext.libraries.append("amdhip64")
                ext.library_dirs = [f"{rocm_path}/lib"]
                ext.include_dirs.append(f"{rocm_path}/include")
                ext.extra_compile_args.append("-D__HIP_PLATFORM_AMD__")
                ext.extra_compile_args.append("-DUSE_ROCM")
                ext.extra_link_args = [f"-L{rocm_path}/lib", "-lamdhip64"]
        else:
            print("=" * 60)
            print("Building for NVIDIA CUDA platform")
            print("=" * 60)

        # Continue with normal build
        build_ext.run(self)


# Detect platform for package_data
platform_type, _, rocm_path_detected = detect_platform()
package_data_patterns = ["*.hpp", "*.h", "cpp.pyi"]

setup(
    packages=[
        "fastsafetensors",
        "fastsafetensors.copier",
        "fastsafetensors.cpp",
        "fastsafetensors.frameworks",
    ],
    include_package_data=True,
    package_data={"fastsafetensors.cpp": package_data_patterns},
    ext_modules=[
        MyExtension(
            name=f"fastsafetensors.cpp",
            sources=["fastsafetensors/cpp/ext.cpp"],
            include_dirs=["fastsafetensors/cpp"],
            mod_name="cpp",
            platform_type=platform_type,
            rocm_path=rocm_path_detected,
        )
    ],
    cmdclass={
        "build_ext": CustomBuildExt,
    },
)



# SPDX-License-Identifier: Apache-2.0

import math
from typing import Any, Dict, List, Optional, OrderedDict, Tuple, Union

from . import cpp as fstcpp
from .common import (
    SafeTensorsMetadata,
    TensorFrame,
    get_device_numa_node,
    init_logger,
    set_debug,
)
from .copier.gds import new_gds_file_copier
from .file_buffer import FilesBufferOnDevice
from .frameworks import TensorBase, get_framework_op
from .st_types import Device, DType
from .tensor_factory import LazyTensorFactory

gl_set_numa = False

loaded_library = False

logger = init_logger(__name__)


class BaseSafeTensorsFileLoader:
    r"""Base class for loading .safetensors files lazily.

    Args:
        pg (Optional[Any]): Process group-like objects for distributed loading.
                           Use None for single device use-cases.
        device (Device): Target device where tensors will be loaded (CPU, CUDA, etc.).
        copier_constructor: Constructor function for creating file copier objects.
        set_numa (bool): Whether to set NUMA node affinity for optimized memory access.
        disable_cache (bool): Whether to disable caching of loaded tensors.
        debug_log (bool): Enable detailed debug logging.
        framework (str): Deep learning framework to use ("pytorch" or "paddle").
    """

    def __init__(
        self,
        pg: Optional[Any],
        device: Device,
        copier_constructor,
        set_numa: bool = True,
        disable_cache: bool = True,
        framework="pytorch",
        **kwargs,
    ):
        self.framework = get_framework_op(framework)
        self.pg = self.framework.get_process_group(pg)
        self.device = device
        self.meta: Dict[str, Tuple[SafeTensorsMetadata, int]] = {}
        self.frames = OrderedDict[str, TensorFrame]()
        self.disable_cache = disable_cache
        self.init_numa(set_numa)
        self.copier_constructor = copier_constructor

    def init_numa(self, set_numa: bool = True):
        global gl_set_numa
        if not gl_set_numa and set_numa:
            node = get_device_numa_node(self.device.index)
            if node is not None:
                fstcpp.set_numa_node(node)
            gl_set_numa = True

    def reset(self):
        self.frames = {}
        self.meta = {}

    def close(self):
        self.reset()
        del self.copier_constructor

    def get_keys(self) -> List[str]:
        return list(self.frames.keys())

    def get_shape(self, tensor_name: str) -> List[int]:
        return self.frames[tensor_name].shape

    def add_filenames(self, filenames: Dict[int, List[str]]):
        """
        Register files to ranks to be copied at copy_file_to_device().
        """
        # shuffle files in a round-robin fashion to avoid OoM
        rank_next_idx = {rank: 0 for rank in filenames.keys()}
        completed = 0
        while completed < len(filenames.keys()):
            completed = 0
            for rank in filenames.keys():
                next_idx = rank_next_idx[rank]
                if next_idx < len(filenames[rank]):
                    realpath = filenames[rank][next_idx]  # os.path.realpath(filename)
                    metadata = SafeTensorsMetadata.from_file(realpath, self.framework)
                    self.meta[realpath] = (metadata, rank)
                    self.frames.update(metadata.tensors)
                    if rank == self.pg.rank():
                        logger.debug(
                            "add_filenames %d: path=%s", len(self.meta), realpath
                        )
                    rank_next_idx[rank] = next_idx + 1
                else:
                    completed += 1

    def copy_files_to_device(
        self,
        dtype: DType = DType.AUTO,
        use_buf_register: bool = True,
        max_copy_block_size: int = 16 * 1024 * 1024 * 1024,
    ) -> FilesBufferOnDevice:
        """
        trigger copying all the files to device buffers.
        At this moment, we do not instantiate tensors but just creating copies at device buffers with or without GDS.
        Users can instantiate and/or partition tensors with FilesBufferOnDevice returned by this function.
        """
        self.framework.set_device(self.device)

        need_wait: List[LazyTensorFactory] = []
        factories: Dict[int, List[LazyTensorFactory]] = {}
        for i in range(0, self.pg.size()):
            factories[i] = []

        factory_idx_bits = math.ceil(math.log2(len(self.meta) + 1))
        lidx = 1
        for _, (meta, rank) in sorted(self.meta.items(), key=lambda x: x[0]):
            copier = self.copier_constructor(meta, self.device, self.framework)
            self_rank = self.pg.rank() == rank
            factory = LazyTensorFactory(
                meta,
                self.device,
                rank,
                self_rank,
                factory_idx_bits,
                lidx,
                copier,
                self.framework,
                disable_cache=self.disable_cache,
            )
            factory.submit_io(use_buf_register, max_copy_block_size)
            factories[rank].append(factory)
            if self_rank:
                need_wait.append(factory)
            lidx += 1
        for factory in need_wait:
            factory.wait_io(dtype=dtype, noalign=False)
        return FilesBufferOnDevice(factories, pg=self.pg, framework=self.framework)


class SafeTensorsFileLoader(BaseSafeTensorsFileLoader):
    r"""Load .safetensors files lazily.

    Args:
        devcie (str): target device.
        pg (Optional[Any]): process group-like objects for distributed. None for single GPU use-cases.
        bbuf_size_kb (int): bounce buffer size for file copies.
        max_threads (int): maximum number of threads for memory copies.
        nogds (bool): if True, trun off GDS and fallback to pread with bounce buffer.
        debug_log (bool): enable debug logs.

    Examples:
        >> from fastsafetensors import SafeTensorsFileLoader
        >> src_files = download(target_dir, "gpt2")
        >> loader = SafeTensorsFileLoader(Device("cpu"), nogds=True, debug_log=True)
        >> loader.add_filenames({0: src_files})
        >> bufs = loader.copy_files_to_device()
        >> print(bufs.get_tensor(loader.get_keys()[0]))
        >> loader.close()
    """

    def __init__(
        self,
        pg: Optional[Any],
        device: str = "cpu",
        bbuf_size_kb: int = 16 * 1024,
        max_threads: int = 16,
        nogds: bool = False,
        set_numa: bool = True,
        disable_cache: bool = True,
        debug_log: bool = False,
        framework="pytorch",
        **kwargs,
    ):
        self.framework = get_framework_op(framework)
        self.pg = self.framework.get_process_group(pg)
        self.device = self.framework.get_device(device, self.pg)

        fstcpp.set_debug_log(debug_log)
        if debug_log:
            set_debug()
        global loaded_library
        if not loaded_library:
            fstcpp.load_library_functions()
            if not nogds:
                # no need to init gds and consume 10s+ in none-gds case
                if fstcpp.init_gds() != 0:
                    raise Exception(f"[FAIL] init_gds()")
            loaded_library = True

        copier = new_gds_file_copier(self.device, bbuf_size_kb, max_threads, nogds)
        super().__init__(
            pg,
            self.device,
            copier,
            set_numa,
            disable_cache,
            framework,
            **kwargs,
        )


class fastsafe_open:
    """
    Opens a safetensors lazily and returns tensors as asked
    This is an enhanced version of safe_open in the original safetensors library to consume file list

    Args:
        filenames (:obj:`str`|`list[str]`|`dict[int, str]`): The filename(s) or rank-file map to open
        framework (:obj:`str`): `pt`, `pytorch`, and `paddle` are only supported currently
        device (:obj:`str`, defaults to :obj:`"cpu"`): The device on which you want the tensors.
    """

    def __init__(
        self,
        filenames: Union[str, List[str], Dict[int, List[str]]],
        framework: str = "pt",
        pg: Optional[Any] = None,
        device: str = "cpu",
        nogds: bool = False,
        debug_log: bool = False,
        max_copy_block_size: int = 16 * 1024 * 1024 * 1024,
    ):
        self.loader = SafeTensorsFileLoader(
            pg, device, nogds=nogds, debug_log=debug_log, framework=framework
        )
        file_dict: Dict[int, List[str]] = {}
        if isinstance(filenames, str):
            file_dict = {0: [filenames]}
        if isinstance(filenames, list):
            file_dict = {0: filenames}
        elif isinstance(filenames, dict):
            file_dict = filenames
        self.loader.add_filenames(file_dict)
        self.fb = self.loader.copy_files_to_device(
            max_copy_block_size=max_copy_block_size
        )

    def metadata(self) -> Dict[str, Dict[str, str]]:
        ret = {}
        for filename, (metadata, _) in self.loader.meta.items():
            ret[filename] = metadata.metadata
        return ret

    def keys(self) -> List[str]:
        return list(self.fb.key_to_rank_lidx.keys())

    def get_tensor_wrapped(self, name: str) -> TensorBase:
        return self.fb.get_tensor_wrapped(name)

    def get_tensor(self, name: str) -> Any:
        return self.get_tensor_wrapped(name).get_raw()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        if self.fb:
            self.fb.close()
        if self.loader:
            self.loader.close()



# SPDX-License-Identifier: Apache-2.0

import warnings
from typing import Dict, Optional

from .. import cpp as fstcpp
from ..common import SafeTensorsMetadata, init_logger, is_gpu_found
from ..frameworks import FrameworkOpBase, TensorBase
from ..st_types import Device, DeviceType, DType
from .base import CopierInterface
from .nogds import NoGdsFileCopier

logger = init_logger(__name__)


class GdsFileCopier(CopierInterface):
    def __init__(
        self,
        metadata: SafeTensorsMetadata,
        device: Device,
        reader: fstcpp.gds_file_reader,
        framework: FrameworkOpBase,
    ):
        self.framework = framework
        self.metadata = metadata
        self.device = device
        self.reader = reader
        self.gbuf = None
        self.fh: Optional[fstcpp.gds_file_handle] = None
        self.copy_reqs: Dict[int, int] = {}
        self.aligned_length = 0
        cuda_ver = framework.get_cuda_ver()
        if cuda_ver and cuda_ver != "0.0":
            # Parse version string (e.g., "cuda-12.1" or "hip-5.7.0")
            # Extract the numeric part after the platform prefix
            ver_parts = cuda_ver.split("-", 1)
            if len(ver_parts) == 2:
                cudavers = list(map(int, ver_parts[1].split(".")))
                # CUDA 12.2 (GDS version 1.7) introduces support for non O_DIRECT file descriptors
                # Compatible with CUDA 11.x
                # Only applies to CUDA platform (not ROCm/HIP)
                if ver_parts[0] == "cuda":
                    self.o_direct = not (
                        cudavers[0] > 12 or (cudavers[0] == 12 and cudavers[1] >= 2)
                    )
                else:
                    # ROCm/HIP platform, use O_DIRECT
                    self.o_direct = True
            else:
                # Fallback if format is unexpected
                self.o_direct = True
        else:
            # No GPU platform detected, use O_DIRECT
            self.o_direct = True

    def set_o_direct(self, enable: bool):
        self.o_direct = enable

    def submit_io(
        self, use_buf_register: bool, max_copy_block_size: int
    ) -> fstcpp.gds_device_buffer:
        dev_is_cuda = (
            self.device.type == DeviceType.CUDA or self.device.type == DeviceType.GPU
        )
        ALIGN: int = fstcpp.get_alignment_size()
        self.fh = fstcpp.gds_file_handle(self.metadata.src, self.o_direct, dev_is_cuda)
        offset = self.metadata.header_length
        length = self.metadata.size_bytes - self.metadata.header_length
        head_bytes = offset % ALIGN
        tail_bytes = (length + head_bytes) % ALIGN
        if tail_bytes > 0:
            tail_bytes = ALIGN - tail_bytes
            aligned_length = length + head_bytes + tail_bytes
        else:
            aligned_length = length + head_bytes
        aligned_offset = offset - head_bytes

        gbuf = self.framework.alloc_tensor_memory(aligned_length, self.device)
        if use_buf_register:
            count = 0
            while count < aligned_length:
                req_len = aligned_length - count
                if req_len > max_copy_block_size:
                    req_len = max_copy_block_size
                if gbuf.cufile_register(count, req_len) < 0:
                    raise Exception(
                        "submit_io: register_buffer failed, ptr=0x{:x}, count={}, len={}".format(
                            gbuf.get_base_address(), count, req_len
                        )
                    )
                count += req_len

        count = 0
        while count < aligned_length:
            req_len = aligned_length - count
            if req_len > max_copy_block_size:
                req_len = max_copy_block_size
            # TODO: pass timeout so that wait_copy_tensors can recognize too slow pread()
            req = self.reader.submit_read(
                self.fh,
                gbuf,
                aligned_offset + count,
                req_len,
                count,
                self.metadata.size_bytes,
            )
            self.copy_reqs[req] = -1 if not use_buf_register else count
            count += req_len
        self.aligned_offset = aligned_offset
        self.aligned_length = aligned_length
        return gbuf

    def wait_io(
        self,
        gbuf: fstcpp.gds_device_buffer,
        dtype: DType = DType.AUTO,
        noalign: bool = False,
    ) -> Dict[str, TensorBase]:
        failed = []
        for req, c in sorted(self.copy_reqs.items(), key=lambda x: x[0]):
            count = self.reader.wait_read(req)
            if count < 0:
                failed.append(req)
            if c != -1:
                gbuf.cufile_deregister(c)
        if self.fh is not None:
            del self.fh
            self.fh = None
        if len(failed) > 0:
            raise Exception(
                f"wait_io: wait_gds_read failed, failed={failed}, reqs={self.copy_reqs}"
            )
        self.copy_reqs = {}
        if not noalign and not self.metadata.aligned and self.aligned_length > 0:
            misaligned_bytes = (
                self.metadata.header_length % self.framework.get_device_ptr_align()
            )
            length = 1024 * 1024 * 1024
            tmp_gbuf = self.framework.alloc_tensor_memory(length, self.device)
            count = 0
            while count + misaligned_bytes < self.aligned_length:
                l = self.aligned_length - misaligned_bytes - count
                if l > length:
                    l = length
                logger.debug(
                    "wait_io: fix misalignment, src=0x%x, misaligned_bytes=%d, count=%d, tmp=0x%x",
                    gbuf.get_base_address(),
                    misaligned_bytes,
                    count,
                    tmp_gbuf.get_base_address(),
                )
                gbuf.memmove(count, misaligned_bytes + count, tmp_gbuf, l)
                count += l
            self.framework.free_tensor_memory(tmp_gbuf, self.device)
            self.aligned_offset += misaligned_bytes
        return self.metadata.get_tensors(
            gbuf, self.device, self.aligned_offset, dtype=dtype
        )


def new_gds_file_copier(
    device: Device,
    bbuf_size_kb: int = 16 * 1024,
    max_threads: int = 16,
    nogds: bool = False,
):
    device_is_not_cpu = device.type != DeviceType.CPU
    if device_is_not_cpu and not is_gpu_found():
        raise Exception(
            "[FAIL] GPU runtime library (libcudart.so or libamdhip64.so) does not exist"
        )
    if device_is_not_cpu and not nogds:
        gds_supported = fstcpp.is_gds_supported(
            device.index if device.index is not None else 0
        )
        if gds_supported < 0:
            raise Exception(f"is_gds_supported({device.index}) failed")
        if not fstcpp.is_cufile_found():
            warnings.warn(
                "libcufile.so does not exist but nogds is False. use nogds=True",
                UserWarning,
            )
            nogds = True
        elif gds_supported == 0:
            warnings.warn(
                "GDS is not supported in this platform but nogds is False. use nogds=True",
                UserWarning,
            )
            nogds = True

    device_id = device.index if device.index is not None else 0
    if nogds:
        nogds_reader = fstcpp.nogds_file_reader(
            False, bbuf_size_kb, max_threads, device_is_not_cpu, device_id
        )

        def construct_nogds_copier(
            metadata: SafeTensorsMetadata,
            device: Device,
            framework: FrameworkOpBase,
        ) -> CopierInterface:
            return NoGdsFileCopier(metadata, device, nogds_reader, framework)

        return construct_nogds_copier

    reader = fstcpp.gds_file_reader(max_threads, device_is_not_cpu, device_id)

    def construct_copier(
        metadata: SafeTensorsMetadata,
        device: Device,
        framework: FrameworkOpBase,
    ) -> CopierInterface:
        return GdsFileCopier(metadata, device, reader, framework)

    return construct_copier



# SPDX-License-Identifier: Apache-2.0

from collections import OrderedDict
from typing import Any, Dict, List, Optional, Tuple

from .common import init_logger
from .frameworks import FrameworkOpBase, ProcessGroupBase, TensorBase
from .st_types import Device, DType
from .tensor_factory import LazyTensorFactory

logger = init_logger(__name__)


class FilesBufferOnDevice:
    r"""Device buffer for .safetensors files.
        Users can call get_tensor(), get_sharded(), etc. to instantiate (sharded) tensors from the device buffer.
        Note that for multi-process loading, users must follow the single-program multiple-data (SPMD) paradigm, which is common for torch.distributed programs.
        In other words, users must ensure that every worker process calls the methods here in the same order.
        This is because methods here reuse torch.distributed operations: broadcast, scatter, recv, and send.
        They synchornously wait all the workers to execute copies among processes.

        Users should create this instance with SafeTensorsFileLoader.submit_io().

    Args:
        rank_loaders (Dict<rank, list(LazyTensorFacotry)>): Tensor factories per rank, which hold device pointers for buffers.
        pg (ProcessGroupBase): process group for calling distributed ops.
        auto_mem_delete (bool): automatically release device buffers when all the tensors are shuffled.

    Examples:
        See examples/run_single.py and examples/run_parallel.py.
    """

    def __init__(
        self,
        rank_loaders: Dict[int, List[LazyTensorFactory]],
        pg: ProcessGroupBase,
        framework: FrameworkOpBase,
        auto_mem_delete: bool = True,
    ):
        self.framework = framework
        self.rank_loaders: Dict[int, List[LazyTensorFactory]] = rank_loaders
        self.key_to_rank_lidx: Dict[str, Tuple[int, int]] = {}
        self.instantiated: Dict[int, Dict[int, Dict[str, bool]]] = {}  # rank, key name
        for rank, loaders in rank_loaders.items():
            self.instantiated[rank] = {}
            for lidx, loader in enumerate(loaders):
                for key in loader.metadata.tensors.keys():
                    if key in self.key_to_rank_lidx:
                        raise Exception(
                            f"FilesBufferOnDevice: key {key} must be unique among files"
                        )
                    self.key_to_rank_lidx[key] = (rank, lidx)
                self.instantiated[rank][lidx] = {}
        self.pg = pg
        self.auto_mem_delete = auto_mem_delete and self.pg.size() > 1

    def close(self):
        for _, loaders in self.rank_loaders.items():
            for loader in loaders:
                loader.free_dev_ptrs()
        self.rank_loaders = {}

    def get_filename(self, tensor_name: str) -> str:
        if tensor_name not in self.key_to_rank_lidx:
            return ""
        rank, lidx = self.key_to_rank_lidx[tensor_name]
        return self.rank_loaders[rank][lidx].metadata.src

    def get_shape(self, tensor_name: str) -> List[int]:
        rank, lidx = self._get_rank_lidx(tensor_name)
        return self.rank_loaders[rank][lidx].metadata.tensors[tensor_name].shape

    def _get_rank_lidx(self, tensor_name: str) -> Tuple[int, int]:
        if tensor_name not in self.key_to_rank_lidx:
            raise ValueError(f"_get_rank: key {tensor_name} was not found in files")
        return self.key_to_rank_lidx[tensor_name]

    def _get_tensor(
        self,
        rank: int,
        lidx: int,
        tensor_name: str,
        ret: TensorBase,
        device: Optional[Device],
        dtype: DType,
    ) -> TensorBase:
        loader = self.rank_loaders[rank][lidx]
        if self.auto_mem_delete:
            self.instantiated[rank][lidx][tensor_name] = True
            if len(self.instantiated[rank][lidx]) == len(loader.metadata.tensors):
                if self.pg.rank() == rank:
                    logger.debug(
                        "_get_tensor: free_dev_ptrs, lidx=%d, src=%s",
                        lidx,
                        loader.metadata.src,
                    )
                loader.free_dev_ptrs()
        return ret.to(device=device, dtype=dtype)

    def get_sharded_wrapped(
        self,
        tensor_name: str,
        dim: int,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> TensorBase:
        rank, lidix = self._get_rank_lidx(tensor_name)
        t = self.rank_loaders[rank][lidix].shuffle(self.pg, tensor_name, dim)
        return self._get_tensor(rank, lidix, tensor_name, t, device, dtype)

    def get_sharded(
        self,
        tensor_name: str,
        dim: int,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> Any:
        """
        partition a tensor instance with the key tensor_name at the dimension dim and return it.
        In multi-process loading, this eventually calls torch.distributed.scatter.
        A special dim is -1, which broadcast a tensor to all the ranks (== get_tensor()).
        """
        return self.get_sharded_wrapped(tensor_name, dim, device, dtype).get_raw()

    def get_tensor_wrapped(
        self,
        tensor_name: str,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> TensorBase:
        return self.get_sharded_wrapped(tensor_name, -1, device, dtype)

    def get_tensor(
        self,
        tensor_name: str,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> Any:
        """
        get a tensor instance with the key tensor_name from a local or remote rank.
        In multi-process loading, this eventually calls torch.distributed.broadcast.
        So, every rank will allocate the same tensor at each device memroy.
        In single-process loading, this directly instantiates a tensor from the device buffer with zero copy.
        """
        return self.get_tensor_wrapped(tensor_name, device, dtype).get_raw()

    def push_tensor(
        self,
        tensor_name: str,
        dst_rank: int,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> Optional[Any]:
        """
        push a tensor instance with the key tensor_name from a rank to a destination rank dst_rank.
        In multi-process loading, this eventually calls torch.distributed.send if the rank has the tensor instance.
        The destination rank will call torch.distributed.recv.
        Other ranks do nothing.
        """
        rank, lidix = self._get_rank_lidx(tensor_name)
        t = self.rank_loaders[rank][lidix].push(self.pg, tensor_name, dst_rank, rank)
        if t:
            return self._get_tensor(
                rank, lidix, tensor_name, t, device, dtype
            ).get_raw()
        return None

    def get_multi_cols(
        self,
        tensor_names: List[str],
        dim: int,
        device: Optional[Device] = None,
        dtype: DType = DType.AUTO,
    ) -> TensorBase:
        rank_lidixs: Dict[Tuple[int, int], List[str]] = {}
        for tensor_name in tensor_names:
            ranklidx = self._get_rank_lidx(tensor_name)
            if ranklidx in rank_lidixs:
                rank_lidixs[ranklidx].append(tensor_name)
            else:
                rank_lidixs[ranklidx] = [tensor_name]
        ts: List[TensorBase] = []
        for (rank, lidix), tns in sorted(rank_lidixs.items(), key=lambda x: x[0]):
            ts.append(
                self.rank_loaders[rank][lidix].shuffle_multi_cols(self.pg, tns, dim)
            )
        if len(ts) == 1:
            # fastpath: tensors at the same layer are often in the same file
            return self._get_tensor(
                rank, lidix, rank_lidixs[(rank, lidix)][0], ts[0], device, dtype
            )
        ret = self.framework.concat_tensors(ts, dim=dim)
        if self.auto_mem_delete:
            for tensor_name in tensor_names:
                rank, lidx = self._get_rank_lidx(tensor_name)
                loader = self.rank_loaders[rank][lidix]
                self.instantiated[rank][lidx][tensor_name] = True
                if len(self.instantiated[rank][lidx]) == len(loader.metadata.tensors):
                    if self.pg.rank() == rank:
                        logger.debug(
                            "get_multi_cols: free_dev_ptrs, rank=%d, lidx=%d, src=%s",
                            rank,
                            lidx,
                            loader.metadata.src,
                        )
                    loader.free_dev_ptrs()
        return ret.to(device=device, dtype=dtype)

    def as_dict(self, tensor_shard_dim: OrderedDict[str, int]) -> Dict[str, TensorBase]:
        tensors: Dict[str, TensorBase] = {}
        for tensor_name, dim in tensor_shard_dim.items():
            rank, lidx = self._get_rank_lidx(tensor_name)
            loader = self.rank_loaders[rank][lidx]
            tensors[tensor_name] = loader.shuffle(self.pg, tensor_name, dim)
            if self.auto_mem_delete:
                self.instantiated[rank][lidx][tensor_name] = True
                if len(self.instantiated[rank][lidx]) == len(loader.metadata.tensors):
                    if self.pg.rank() == rank:
                        logger.debug(
                            "as_dict: free_dev_ptrs, rank=%d, src=%s",
                            rank,
                            loader.metadata.src,
                        )
                    loader.free_dev_ptrs()
        if self.auto_mem_delete:
            self.rank_loaders = {}
        return tensors
